from functools import partial

import pytest
from plumbum import ProcessExecutionError
from plumbum.cmd import pg_dump

from pg_grant import parse_acl_item, FunctionInfo, PgObjectType
from pg_grant.query import (
    get_all_function_acls, get_all_sequence_acls, get_all_table_acls,
    get_all_type_acls)


def _priv_acls(conn, acls, type_, revoke):
    for obj in acls:
        arg_types = None
        if isinstance(obj, FunctionInfo):
            arg_types = obj.arg_types

        if obj.acl is not None:
            for acl in obj.acl:
                parsed = parse_acl_item(acl)

                if revoke:
                    statements = parsed.as_revoke_statements(
                        type_, obj.name, schema=obj.schema, arg_types=arg_types)
                else:
                    statements = parsed.as_grant_statements(
                        type_, obj.name, schema=obj.schema, arg_types=arg_types)

                for stmt in statements:
                    conn.execute(stmt)


grant_table_acls = partial(_priv_acls, type_=PgObjectType.TABLE, revoke=False)
revoke_table_acls = partial(_priv_acls, type_=PgObjectType.TABLE, revoke=True)
grant_sequence_acls = partial(_priv_acls, type_=PgObjectType.SEQUENCE, revoke=False)
revoke_sequence_acls = partial(_priv_acls, type_=PgObjectType.SEQUENCE, revoke=True)
grant_type_acls = partial(_priv_acls, type_=PgObjectType.TYPE, revoke=False)
revoke_type_acls = partial(_priv_acls, type_=PgObjectType.TYPE, revoke=True)
grant_function_acls = partial(_priv_acls, type_=PgObjectType.FUNCTION, revoke=False)
revoke_function_acls = partial(_priv_acls, type_=PgObjectType.FUNCTION, revoke=True)


@pytest.mark.parametrize('get, revoke, grant', [
    (get_all_table_acls, revoke_table_acls, grant_table_acls),
    (get_all_sequence_acls, revoke_sequence_acls, grant_sequence_acls),
    (get_all_type_acls, revoke_type_acls, grant_type_acls),
    (get_all_function_acls, revoke_function_acls, grant_function_acls),
])
def test_revoke_grant_schema_relations(connection, postgres_url, get, revoke, grant):
    cmd = pg_dump['--schema-only', postgres_url]

    try:
        code, dump1, err = cmd.run()
    except ProcessExecutionError as exc:
        if 'server version mismatch' in exc.stderr:
            pytest.skip("pg_dump is older than server version")
        else:
            raise

    assert code == 0

    with connection.begin():
        acls = get(connection, 'public')
        revoke(connection, acls)

    code, dump2, _ = cmd.run()
    assert code == 0
    assert dump1 != dump2

    with connection.begin():
        grant(connection, acls)

    code, dump3, _ = cmd.run()
    assert code == 0

    assert dump1 == dump3



import pytest

from pg_grant import NoSuchObjectError
from pg_grant.query import get_all_type_acls, get_type_acl


expected_acls = {
    'public': {
        # bug_status has default privileges, so None is returned.
        'bug_status': None,
        # alice is owner, bob was granted usage, public must get usage by default
        # https://stackoverflow.com/questions/46656644
        'thing': ['=U/alice', 'alice=U/alice', 'bob=U/alice'],
    },
}


@pytest.mark.parametrize('name, acls', expected_acls['public'].items())
def test_get_type_acl_visible(connection, name, acls):
    """Find visible (i.e. in search path) types matching ``name``."""
    type_ = get_type_acl(connection, name)
    assert type_.acl == acls


@pytest.mark.parametrize('schema, name, acls', [
    (schema, name, acl)
    for schema, d in expected_acls.items()
    for name, acl in d.items()
])
def test_get_type_acl_schema(connection, schema, name, acls):
    """Find types  from ``schema`` matching ``name``."""
    type_ = get_type_acl(connection, name, schema)
    assert type_.acl == acls


def test_get_all_type_acls(connection):
    """Get all sequences in all schemas."""
    types = get_all_type_acls(connection)
    schemas = {x.schema for x in types}
    assert schemas >= {'public', 'information_schema', 'pg_catalog'}

    tested = 0

    for type_ in types:
        if type_.schema not in expected_acls:
            continue

        if type_.name not in expected_acls[type_.schema]:
            continue

        assert type_.acl == expected_acls[type_.schema][type_.name]
        tested += 1

    assert tested == sum(len(v) for v in expected_acls.values())


def test_no_such_object(connection):
    with pytest.raises(NoSuchObjectError):
        get_type_acl(connection, 'db2')



import os
from contextlib import closing
from itertools import chain
from pathlib import Path

import psycopg2
from psycopg2 import sql
import pytest
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT
from sqlalchemy import create_engine, text
from sqlalchemy.engine.url import make_url

tests_dir = Path(__file__).parents[0].resolve()
test_schema_file = Path(tests_dir, 'data', 'test-schema.sql')

# This matches docker-compose.yml for easy local development
DEFAULT_DATABASE_URL = 'postgresql://postgres@127.0.0.1:5440/postgres'


@pytest.fixture(scope='session')
def postgres_url():
    # The only global state we want from the system is a superuser connection to
    # an empty database cluster.
    superuser_url = make_url(os.getenv("DATABASE_URL", DEFAULT_DATABASE_URL))

    dbname = 'db1'
    superusers = ['alice']
    users = ['bob', 'charlie']
    password = 'hunter2'

    # We'll make a new database and roles in the cluster, so let's check that
    # they don't match that of the configured superuser url.
    assert superuser_url.username not in {*superusers, *users}
    assert superuser_url.database != dbname

    conn = psycopg2.connect(superuser_url.render_as_string(hide_password=False))
    with closing(conn), conn.cursor() as cur:
        conn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
        for user in superusers:
            stmt = sql.SQL("CREATE USER {} WITH SUPERUSER PASSWORD {}")
            cur.execute(stmt.format(sql.Identifier(user), sql.Literal(password)))
        for user in users:
            stmt = sql.SQL("CREATE USER {}")
            cur.execute(stmt.format(sql.Identifier(user)))
        cur.execute(sql.SQL("CREATE DATABASE {}").format(sql.Identifier(dbname)))

    test_url = superuser_url.set(database=dbname, username=superusers[0], password=password)

    yield test_url.render_as_string(hide_password=False)

    conn = psycopg2.connect(superuser_url.render_as_string(hide_password=False))
    with closing(conn), conn.cursor() as cur:
        conn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
        cur.execute(sql.SQL("DROP DATABASE {}").format(sql.Identifier(dbname)))
        for user in chain(superusers, users):
            user_ident = sql.Identifier(user)
            cur.execute(sql.SQL("DROP OWNED BY {} CASCADE").format(user_ident))
            cur.execute(sql.SQL("DROP USER {}").format(user_ident))


@pytest.fixture(scope='session')
def engine(postgres_url):
    engine = create_engine(postgres_url)
    yield engine
    engine.dispose()


@pytest.fixture(scope='session')
def pg_schema(engine):
    with test_schema_file.open() as fp:
        with engine.begin() as conn:
            conn.execute(text(fp.read()))


@pytest.fixture
def connection(engine, pg_schema):
    with engine.connect() as conn:
        yield conn



from unittest.mock import patch

import pytest
from sqlalchemy import Column, Integer, MetaData, Sequence, Table, table
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import declarative_base

from pg_grant import PgObjectType, Privileges
from pg_grant.sql import revoke

meta = MetaData()
Base = declarative_base()

simple_table = table('users')
simple_table_kw = table('user')
full_table = Table('users', meta)
full_table_kw = Table('user', meta)
full_table_schema = Table('users', meta, schema='s')
seq = Sequence('user_id')
seq_kw = Sequence('user')
seq_schema = Sequence('user_id', schema='s')


class ModelA(Base):
    __tablename__ = 'user'
    id = Column(Integer, primary_key=True)


class ModelB(Base):
    __tablename__ = 'users'
    __table_args__ = {'schema': 's'}
    id = Column(Integer, primary_key=True)


@pytest.mark.parametrize('privs, grantee, compiled', [
    (['ALL'], 'alice', 'REVOKE ALL ON TABLE mytable FROM alice'),
    (['ALL'], 'PUBLIC', 'REVOKE ALL ON TABLE mytable FROM PUBLIC'),
    (['ALL'], 'public', 'REVOKE ALL ON TABLE mytable FROM public'),
    (['ALL'], 'grant', 'REVOKE ALL ON TABLE mytable FROM "grant"'),
    (['SELECT'], 'alice', 'REVOKE SELECT ON TABLE mytable FROM alice'),
    (['SELECT', 'INSERT'], 'alice', 'REVOKE SELECT, INSERT ON TABLE mytable FROM alice'),
])
def test_compile_revoke_table_privs(privs, grantee, compiled):
    statement = revoke(privs, PgObjectType.TABLE, 'mytable', grantee)
    assert str(statement.compile(dialect=postgresql.dialect())) == compiled


@pytest.mark.parametrize('target, schema, compiled', [
    ('t', None, 'REVOKE ALL ON TABLE t FROM alice'),
    ('user', None, 'REVOKE ALL ON TABLE "user" FROM alice'),
    ('t', 's', 'REVOKE ALL ON TABLE s.t FROM alice'),
    ('t', 'user', 'REVOKE ALL ON TABLE "user".t FROM alice'),
    (simple_table, None, 'REVOKE ALL ON TABLE users FROM alice'),
    (simple_table_kw, None, 'REVOKE ALL ON TABLE "user" FROM alice'),
    (full_table, None, 'REVOKE ALL ON TABLE users FROM alice'),
    (full_table_kw, None, 'REVOKE ALL ON TABLE "user" FROM alice'),
    (full_table_schema, None, 'REVOKE ALL ON TABLE s.users FROM alice'),
    (ModelA, None, 'REVOKE ALL ON TABLE "user" FROM alice'),
    (ModelB, None, 'REVOKE ALL ON TABLE s.users FROM alice'),
])
def test_compile_revoke_table_target(target, schema, compiled):
    statement = revoke(['ALL'], PgObjectType.TABLE, target, 'alice', schema=schema)
    assert str(statement.compile(dialect=postgresql.dialect())) == compiled


@pytest.mark.parametrize('revoke_option, compiled', [
    (True, 'REVOKE GRANT OPTION FOR ALL ON TABLE t FROM alice'),
    (False, 'REVOKE ALL ON TABLE t FROM alice'),
])
def test_compile_revoke_table_revoke_option(revoke_option, compiled):
    statement = revoke(['ALL'], PgObjectType.TABLE, 't', 'alice', revoke_option)
    assert str(statement.compile(dialect=postgresql.dialect())) == compiled


@pytest.mark.parametrize('target, schema, compiled', [
    ('t', None, 'REVOKE ALL ON SEQUENCE t FROM alice'),
    ('user', None, 'REVOKE ALL ON SEQUENCE "user" FROM alice'),
    ('t', 's', 'REVOKE ALL ON SEQUENCE s.t FROM alice'),
    ('t', 'user', 'REVOKE ALL ON SEQUENCE "user".t FROM alice'),
    (seq, None, 'REVOKE ALL ON SEQUENCE user_id FROM alice'),
    (seq_kw, None, 'REVOKE ALL ON SEQUENCE "user" FROM alice'),
    (seq_schema, None, 'REVOKE ALL ON SEQUENCE s.user_id FROM alice'),
])
def test_compile_revoke_sequence_target(target, schema, compiled):
    statement = revoke(['ALL'], PgObjectType.SEQUENCE, target, 'alice', schema=schema)
    assert str(statement.compile(dialect=postgresql.dialect())) == compiled


@pytest.mark.parametrize('target, schema, compiled', [
    ('t', None, 'REVOKE ALL ON TYPE t FROM alice'),
    ('user', None, 'REVOKE ALL ON TYPE "user" FROM alice'),
    ('t', 's', 'REVOKE ALL ON TYPE s.t FROM alice'),
    ('t', 'user', 'REVOKE ALL ON TYPE "user".t FROM alice'),
])
def test_compile_revoke_type_target(target, schema, compiled):
    statement = revoke(['ALL'], PgObjectType.TYPE, target, 'alice', schema=schema)
    assert str(statement.compile(dialect=postgresql.dialect())) == compiled


@pytest.mark.parametrize('target, arg_types, schema, compiled', [
    ('t', (), None, 'REVOKE ALL ON FUNCTION t() FROM alice'),
    ('user', (), None, 'REVOKE ALL ON FUNCTION "user"() FROM alice'),
    ('t', (), 's', 'REVOKE ALL ON FUNCTION s.t() FROM alice'),
    ('t', (), 'grant', 'REVOKE ALL ON FUNCTION "grant".t() FROM alice'),
    ('t', ('integer',), None, 'REVOKE ALL ON FUNCTION t(integer) FROM alice'),
    ('user', ('text', 'integer'), None,
     'REVOKE ALL ON FUNCTION "user"(text, integer) FROM alice'),
])
def test_compile_revoke_function_target(target, arg_types, schema, compiled):
    statement = revoke(
        ['ALL'], PgObjectType.FUNCTION, target, 'alice', schema=schema,
        arg_types=arg_types)
    assert str(statement.compile(dialect=postgresql.dialect())) == compiled


@pytest.mark.parametrize('type, target, compiled', [
    (PgObjectType.LANGUAGE, 't', 'REVOKE ALL ON LANGUAGE t FROM alice'),
    (PgObjectType.LANGUAGE, 'user', 'REVOKE ALL ON LANGUAGE "user" FROM alice'),
    (PgObjectType.SCHEMA, 't', 'REVOKE ALL ON SCHEMA t FROM alice'),
    (PgObjectType.SCHEMA, 'user', 'REVOKE ALL ON SCHEMA "user" FROM alice'),
    (PgObjectType.DATABASE, 't', 'REVOKE ALL ON DATABASE t FROM alice'),
    (PgObjectType.DATABASE, 'user', 'REVOKE ALL ON DATABASE "user" FROM alice'),
    (PgObjectType.TABLESPACE, 't', 'REVOKE ALL ON TABLESPACE t FROM alice'),
    (PgObjectType.TABLESPACE, 'user', 'REVOKE ALL ON TABLESPACE "user" FROM alice'),
])
def test_compile_revoke_other_target(type, target, compiled):
    statement = revoke(['ALL'], type, target, 'alice')
    assert str(statement.compile(dialect=postgresql.dialect())) == compiled


@pytest.mark.parametrize('type, target, schema, arg_types', [
    (PgObjectType.TABLE, simple_table, 's', None),
    (PgObjectType.SEQUENCE, seq, 's', None),
    ('blah', 't', 's', None),
    (PgObjectType.SEQUENCE, 't', 's', ()),
    (PgObjectType.SEQUENCE, 't', 's', ('integer',)),
])
def test_compile_revoke_invalid(type, target, schema, arg_types):
    statement = revoke(
        ['ALL'], type, target, 'alice', schema=schema, arg_types=arg_types)
    with pytest.raises(ValueError):
        str(statement.compile(dialect=postgresql.dialect()))


@pytest.mark.parametrize('privs', [
    ['BLAH', 'SELECT'],
    ['BLAH'],
])
def test_compile_revoke_privs_invalid(privs):
    with pytest.raises(ValueError):
        str(revoke(privs, PgObjectType.TABLE, 't', 'alice'))


@pytest.mark.parametrize('grantee, privs, privswgo, type_, target, kw', [
    ('alice', ['SELECT', 'INSERT'], [], PgObjectType.FOREIGN_TABLE, 'table1', {}),
    ('alice', ['SELECT'], ['INSERT'], PgObjectType.TABLE, 'table2', {'schema': 's'}),
])
def test_privileges_as_revoke_statements(grantee, privs, privswgo, type_, target, kw):
    priv = Privileges(grantee, 'bob', privs, privswgo)

    def fake_revoke(*args, **kwargs):
        return args, kwargs

    patch_revoke = patch('pg_grant.sql.revoke', fake_revoke)

    with patch_revoke:
        r = priv.as_revoke_statements(type_, target, **kw)

    expected = []

    if privs:
        expected.append(((privs, type_, target, grantee), kw))

    if privswgo:
        expected.append(((privswgo, type_, target, grantee), kw))

    assert r == expected



from .exc import NoSuchObjectError
from .parse import get_default_privileges, parse_acl, parse_acl_item
from .types import (
    FunctionInfo, PgObjectType, Privileges, RelationInfo, SchemaRelationInfo)

__version__ = '0.4.0'
__description__ = 'Parse PostgreSQL privileges'

__license__ = 'MIT'

__author__ = 'Frazer McLean'
__email__ = 'frazer@frazermclean.co.uk'

__all__ = (
    'FunctionInfo',
    'NoSuchObjectError',
    'PgObjectType',
    'Privileges',
    'RelationInfo',
    'SchemaRelationInfo',
    'get_default_privileges',
    'parse_acl',
    'parse_acl_item',
)



from enum import Enum

import attr


class PgObjectType(Enum):
    """PostgreSQL object type."""
    TABLE = 'TABLE'
    SEQUENCE = 'SEQUENCE'
    FUNCTION = 'FUNCTION'
    LANGUAGE = 'LANGUAGE'
    SCHEMA = 'SCHEMA'
    DATABASE = 'DATABASE'
    TABLESPACE = 'TABLESPACE'
    TYPE = 'TYPE'
    FOREIGN_DATA_WRAPPER = 'FOREIGN DATA WRAPPER'
    FOREIGN_SERVER = 'FOREIGN SERVER'
    FOREIGN_TABLE = 'FOREIGN TABLE'
    LARGE_OBJECT = 'LARGE OBJECT'


@attr.s(slots=True)
class Privileges:
    """Stores information from a parsed privilege string.

    .. seealso:: :func:`~.parse.parse_acl_item`
    """
    grantee = attr.ib()
    grantor = attr.ib()
    privs = attr.ib(factory=list)
    privswgo = attr.ib(factory=list)

    def as_grant_statements(self, type_: PgObjectType, target, **kwargs):
        """Return array of :func:`~.sql.grant` statements that can be executed
        to grant these privileges. Refer to the function documentation for the
        meaning of `target` and additional keyword arguments.

        .. note:: This requires installing with the ``sqlalchemy`` extra.
        """
        from .sql import grant

        statements = []

        if self.privs:
            statements.append(
                grant(self.privs, type_, target, self.grantee, **kwargs))

        if self.privswgo:
            statements.append(grant(
                self.privswgo, type_, target, self.grantee, grant_option=True,
                **kwargs))

        return statements

    def as_revoke_statements(self, type_: PgObjectType, target, **kwargs):
        """Return array of :func:`~.sql.revoke` statements that can be executed
        to revoke these privileges. Refer to the function documentation for the
        meaning of `target` and additional keyword arguments.

        .. note::

            The statement for the `privswgo` privileges will revoke them
            fully, not only their grant options.

        .. note:: This requires installing with the ``sqlalchemy`` extra.
        """
        from .sql import revoke

        statements = []

        if self.privs:
            statements.append(
                revoke(self.privs, type_, target, self.grantee, **kwargs))

        if self.privswgo:
            statements.append(revoke(
                self.privswgo, type_, target, self.grantee, **kwargs))

        return statements


@attr.s(slots=True)
class RelationInfo:
    """Holds object information and privileges as queried using the
    :mod:`.query` submodule."""
    #: Row identifier.
    oid = attr.ib()

    #: Name of the table, sequence, etc.
    name = attr.ib()

    #: Owner of the relation.
    owner = attr.ib()

    #: Access control list.
    acl = attr.ib()


@attr.s(slots=True)
class SchemaRelationInfo(RelationInfo):
    """Holds object information and privileges as queried using the
    :mod:`.query` submodule."""
    #: The name of the schema that contains this relation.
    schema = attr.ib()


@attr.s(slots=True)
class FunctionInfo(SchemaRelationInfo):
    """Holds object information and privileges as queried using the
    :mod:`.query` submodule."""
    #: Data types of the function arguments.
    arg_types = attr.ib()


@attr.s(slots=True)
class ColumnInfo:
    """Holds object information and privileges as queried using the
    :mod:`.query` submodule."""
    #: Table identifier.
    table_oid = attr.ib()

    #: The name of the schema that contains the table.
    schema = attr.ib()

    #: Name of the table.
    table = attr.ib()

    #: Name of the column.
    column = attr.ib()

    #: Owner of the table.
    owner = attr.ib()

    #: Column access control list.
    acl = attr.ib()



from unittest.mock import patch

import pytest
from sqlalchemy import Column, Integer, MetaData, Sequence, Table, table
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import declarative_base

from pg_grant import PgObjectType, Privileges
from pg_grant.sql import grant

meta = MetaData()
Base = declarative_base()

simple_table = table('users')
simple_table_kw = table('user')
full_table = Table('users', meta)
full_table_kw = Table('user', meta)
full_table_schema = Table('users', meta, schema='s')
seq = Sequence('user_id')
seq_kw = Sequence('user')
seq_schema = Sequence('user_id', schema='s')


class ModelA(Base):
    __tablename__ = 'user'
    id = Column(Integer, primary_key=True)


class ModelB(Base):
    __tablename__ = 'users'
    __table_args__ = {'schema': 's'}
    id = Column(Integer, primary_key=True)


@pytest.mark.parametrize('privs, grantee, compiled', [
    (['ALL'], 'alice', 'GRANT ALL ON TABLE mytable TO alice'),
    ('ALL', 'PUBLIC', 'GRANT ALL ON TABLE mytable TO PUBLIC'),
    (['ALL'], 'public', 'GRANT ALL ON TABLE mytable TO public'),
    (['ALL'], 'grant', 'GRANT ALL ON TABLE mytable TO "grant"'),
    (['SELECT'], 'alice', 'GRANT SELECT ON TABLE mytable TO alice'),
    (['SELECT', 'INSERT'], 'alice', 'GRANT SELECT, INSERT ON TABLE mytable TO alice'),
])
def test_compile_grant_table_privs(privs, grantee, compiled):
    statement = grant(privs, PgObjectType.TABLE, 'mytable', grantee)
    assert str(statement.compile(dialect=postgresql.dialect())) == compiled


@pytest.mark.parametrize('target, schema, compiled', [
    ('t', None, 'GRANT ALL ON TABLE t TO alice'),
    ('user', None, 'GRANT ALL ON TABLE "user" TO alice'),
    ('t', 's', 'GRANT ALL ON TABLE s.t TO alice'),
    ('t', 'grant', 'GRANT ALL ON TABLE "grant".t TO alice'),
    (simple_table, None, 'GRANT ALL ON TABLE users TO alice'),
    (simple_table_kw, None, 'GRANT ALL ON TABLE "user" TO alice'),
    (full_table, None, 'GRANT ALL ON TABLE users TO alice'),
    (full_table_kw, None, 'GRANT ALL ON TABLE "user" TO alice'),
    (full_table_schema, None, 'GRANT ALL ON TABLE s.users TO alice'),
    (ModelA, None, 'GRANT ALL ON TABLE "user" TO alice'),
    (ModelB, None, 'GRANT ALL ON TABLE s.users TO alice'),
])
def test_compile_grant_table_target(target, schema, compiled):
    statement = grant(
        ['ALL'], PgObjectType.TABLE, target, 'alice', schema=schema)
    assert str(statement.compile(dialect=postgresql.dialect())) == compiled


@pytest.mark.parametrize('grant_option, compiled', [
    (True, 'GRANT ALL ON TABLE t TO alice WITH GRANT OPTION'),
    (False, 'GRANT ALL ON TABLE t TO alice'),
])
def test_compile_grant_table_grant_option(grant_option, compiled):
    statement = grant(['ALL'], PgObjectType.TABLE, 't', 'alice', grant_option)
    assert str(statement.compile(dialect=postgresql.dialect())) == compiled


@pytest.mark.parametrize('target, schema, compiled', [
    ('t', None, 'GRANT ALL (a) ON TABLE t TO alice'),
    ('user', None, 'GRANT ALL (a) ON TABLE "user" TO alice'),
    ('t', 's', 'GRANT ALL (a) ON TABLE s.t TO alice'),
    ('t', 'grant', 'GRANT ALL (a) ON TABLE "grant".t TO alice'),
    (simple_table, None, 'GRANT ALL (a) ON TABLE users TO alice'),
    (simple_table_kw, None, 'GRANT ALL (a) ON TABLE "user" TO alice'),
    (full_table, None, 'GRANT ALL (a) ON TABLE users TO alice'),
    (full_table_kw, None, 'GRANT ALL (a) ON TABLE "user" TO alice'),
    (full_table_schema, None, 'GRANT ALL (a) ON TABLE s.users TO alice'),
    (ModelA, None, 'GRANT ALL (a) ON TABLE "user" TO alice'),
    (ModelB, None, 'GRANT ALL (a) ON TABLE s.users TO alice'),
])
def test_compile_grant_table_column(target, schema, compiled):
    statement = grant(
        ['ALL (a)'], PgObjectType.TABLE, target, 'alice', schema=schema)
    assert str(statement.compile(dialect=postgresql.dialect())) == compiled


@pytest.mark.parametrize('target, schema, compiled', [
    ('t', None, 'GRANT ALL ("user") ON TABLE t TO alice'),
])
def test_compile_grant_table_column_quote(target, schema, compiled):
    statement = grant(
        ['ALL (user)'], PgObjectType.TABLE, target, 'alice', schema=schema)
    assert str(statement.compile(dialect=postgresql.dialect())) == compiled


@pytest.mark.parametrize('target, schema, compiled', [
    ('t', None, 'GRANT ALL ("user") ON TABLE t TO alice'),
])
def test_compile_grant_table_column_quoted_already(target, schema, compiled):
    statement = grant(
        ['ALL ("user")'], PgObjectType.TABLE, target, 'alice', schema=schema,
        quote_subname=False)
    assert str(statement.compile(dialect=postgresql.dialect())) == compiled


@pytest.mark.parametrize('target, schema, compiled', [
    ('t', None, 'GRANT ALL ON SEQUENCE t TO alice'),
    ('user', None, 'GRANT ALL ON SEQUENCE "user" TO alice'),
    ('t', 's', 'GRANT ALL ON SEQUENCE s.t TO alice'),
    ('t', 'grant', 'GRANT ALL ON SEQUENCE "grant".t TO alice'),
    (seq, None, 'GRANT ALL ON SEQUENCE user_id TO alice'),
    (seq_kw, None, 'GRANT ALL ON SEQUENCE "user" TO alice'),
    (seq_schema, None, 'GRANT ALL ON SEQUENCE s.user_id TO alice'),
])
def test_compile_grant_sequence_target(target, schema, compiled):
    statement = grant(
        ['ALL'], PgObjectType.SEQUENCE, target, 'alice', schema=schema)
    assert str(statement.compile(dialect=postgresql.dialect())) == compiled


@pytest.mark.parametrize('target, schema, compiled', [
    ('t', None, 'GRANT ALL ON TYPE t TO alice'),
    ('user', None, 'GRANT ALL ON TYPE "user" TO alice'),
    ('t', 's', 'GRANT ALL ON TYPE s.t TO alice'),
    ('t', 'grant', 'GRANT ALL ON TYPE "grant".t TO alice'),
])
def test_compile_grant_type_target(target, schema, compiled):
    statement = grant(['ALL'], PgObjectType.TYPE, target, 'alice', schema=schema)
    assert str(statement.compile(dialect=postgresql.dialect())) == compiled


@pytest.mark.parametrize('target, arg_types, schema, compiled', [
    ('t', (), None, 'GRANT ALL ON FUNCTION t() TO alice'),
    ('user', (), None, 'GRANT ALL ON FUNCTION "user"() TO alice'),
    ('t', (), 's', 'GRANT ALL ON FUNCTION s.t() TO alice'),
    ('t', (), 'grant', 'GRANT ALL ON FUNCTION "grant".t() TO alice'),
    ('t', ('integer',), None, 'GRANT ALL ON FUNCTION t(integer) TO alice'),
    ('user', ('text', 'integer'), None,
     'GRANT ALL ON FUNCTION "user"(text, integer) TO alice'),
])
def test_compile_grant_function_target(target, arg_types, schema, compiled):
    statement = grant(
        ['ALL'], PgObjectType.FUNCTION, target, 'alice', schema=schema,
        arg_types=arg_types)
    assert str(statement.compile(dialect=postgresql.dialect())) == compiled


@pytest.mark.parametrize('type, target, compiled', [
    (PgObjectType.LANGUAGE, 't', 'GRANT ALL ON LANGUAGE t TO alice'),
    (PgObjectType.LANGUAGE, 'user', 'GRANT ALL ON LANGUAGE "user" TO alice'),
    (PgObjectType.SCHEMA, 't', 'GRANT ALL ON SCHEMA t TO alice'),
    (PgObjectType.SCHEMA, 'user', 'GRANT ALL ON SCHEMA "user" TO alice'),
    (PgObjectType.DATABASE, 't', 'GRANT ALL ON DATABASE t TO alice'),
    (PgObjectType.DATABASE, 'user', 'GRANT ALL ON DATABASE "user" TO alice'),
    (PgObjectType.TABLESPACE, 't', 'GRANT ALL ON TABLESPACE t TO alice'),
    (PgObjectType.TABLESPACE, 'user', 'GRANT ALL ON TABLESPACE "user" TO alice'),
])
def test_compile_grant_other_target(type, target, compiled):
    statement = grant(['ALL'], type, target, 'alice')
    assert str(statement.compile(dialect=postgresql.dialect())) == compiled


@pytest.mark.parametrize('type, target, schema, arg_types', [
    (PgObjectType.TABLE, simple_table, 's', None),
    (PgObjectType.SEQUENCE, seq, 's', None),
    ('blah', 't', 's', None),
    (PgObjectType.SEQUENCE, 't', 's', ()),
    (PgObjectType.SEQUENCE, 't', 's', ('integer',)),
])
def test_compile_grant_invalid(type, target, schema, arg_types):
    statement = grant(
        ['ALL'], type, target, 'alice', schema=schema, arg_types=arg_types)
    with pytest.raises(ValueError):
        str(statement.compile(dialect=postgresql.dialect()))


@pytest.mark.parametrize('privs', [
    ['BLAH', 'SELECT'],
    ['BLAH'],
])
def test_compile_grant_privs_invalid(privs):
    with pytest.raises(ValueError):
        str(grant(privs, PgObjectType.TABLE, 't', 'alice'))


def test_compile_grant_function_privs_invalid():
    with pytest.raises(ValueError) as exc_info:
        str(grant(['ALL'], PgObjectType.FUNCTION, 'fun1', 'alice'))
    assert 'use an empty sequence' in exc_info.value.args[0]


@pytest.mark.parametrize('grantee, privs, privswgo, type_, target, kw', [
    ('alice', ['SELECT', 'INSERT'], [], PgObjectType.FOREIGN_TABLE, 'table1', {}),
    ('alice', ['SELECT'], ['INSERT'], PgObjectType.TABLE, 'table2', {'schema': 's'}),
])
def test_privileges_as_grant_statements(grantee, privs, privswgo, type_, target, kw):
    priv = Privileges(grantee, 'bob', privs, privswgo)

    def fake_grant(*args, **kwargs):
        return args, kwargs

    patch_grant = patch('pg_grant.sql.grant', fake_grant)

    with patch_grant:
        r = priv.as_grant_statements(type_, target, **kw)

    expected = []

    if privs:
        expected.append(((privs, type_, target, grantee), kw))

    if privswgo:
        gokw = kw.copy()
        gokw.update(grant_option=True)
        expected.append(((privswgo, type_, target, grantee), gokw))

    assert r == expected



import pytest

from pg_grant import NoSuchObjectError
from pg_grant.query import get_all_schema_acls, get_schema_acl


def test_get_schema_acl(connection, expected_acls):
    """Find visible schemas matching ``name``."""
    for name, acls in expected_acls.items():
        schema = get_schema_acl(connection, name)
        assert schema.acl == acls


@pytest.fixture
def expected_acls(connection, request):
    if connection.connection.server_version >= 150000:
        return {
            # pg_database_owner is owner, public get usage on the public schema
            # by default
            'public': ['pg_database_owner=UC/pg_database_owner', '=U/pg_database_owner'],
            # schema1 has default privileges, so None is returned.
            'schema1': None,
        }
    else:
        return {
            # postgres is owner, public get access to the public schema by
            # default
            'public': ['postgres=UC/postgres', '=UC/postgres'],
            # schema1 has default privileges, so None is returned.
            'schema1': None,
        }


def test_get_all_schema_acls(connection, expected_acls):
    """Get all schemas in all schemas."""
    schemas = get_all_schema_acls(connection)

    tested = 0

    for schema in schemas:
        if schema.name not in expected_acls:
            continue

        assert schema.acl == expected_acls[schema.name]
        tested += 1

    assert tested == len(expected_acls)


def test_no_such_object(connection):
    with pytest.raises(NoSuchObjectError):
        get_schema_acl(connection, 'schema2')



import pytest

from pg_grant import NoSuchObjectError
from pg_grant.query import get_all_language_acls, get_language_acl


expected_acls = {
    # sql has default privileges, so None is returned.
    'sql': None,
    # postgres is owner, alice was granted usage
    'plpgsql': ['postgres=U/postgres', 'alice=U/postgres'],
}


@pytest.mark.parametrize('name, acls', expected_acls.items())
def test_get_language_acl(connection, name, acls):
    """Find visible languages matching ``name``."""
    language = get_language_acl(connection, name)
    assert language.acl == acls


def test_get_all_language_acls(connection):
    """Get all languages in all schemas."""
    languages = get_all_language_acls(connection)

    tested = 0

    for language in languages:
        if language.name not in expected_acls:
            continue

        assert language.acl == expected_acls[language.name]
        tested += 1

    assert tested == len(expected_acls)


def test_no_such_object(connection):
    with pytest.raises(NoSuchObjectError):
        get_language_acl(connection, 'eggs')



from enum import Enum
from typing import Sequence

from sqlalchemy import ARRAY, Text, cast, column, func, select, table, text
from sqlalchemy.engine import Connectable

from .exc import NoSuchObjectError
from .types import ColumnInfo, FunctionInfo, RelationInfo, SchemaRelationInfo

__all__ = (
    'get_all_table_acls',
    'get_table_acl',
    'get_all_column_acls',
    'get_column_acls',
    'get_all_sequence_acls',
    'get_sequence_acl',
    'get_all_function_acls',
    'get_function_acl',
    'get_all_language_acls',
    'get_language_acl',
    'get_all_schema_acls',
    'get_schema_acl',
    'get_all_database_acls',
    'get_database_acl',
    'get_all_tablespace_acls',
    'get_tablespace_acl',
    'get_all_type_acls',
    'get_type_acl',
)

pg_table_is_visible = func.pg_catalog.pg_table_is_visible
pg_function_is_visible = func.pg_catalog.pg_function_is_visible
pg_type_is_visible = func.pg_catalog.pg_type_is_visible
array_agg = func.array_agg
unnest = func.unnest
coalesce = func.coalesce
canonical_type = func.pg_temp.pg_grant_canonical_type


class PgRelKind(Enum):
    TABLE = 'r'
    INDEX = 'i'
    SEQUENCE = 'S'
    VIEW = 'v'
    MATERIALIZED_VIEW = 'm'
    COMPOSITE_TYPE = 'c'
    TOAST_TABLE = 't'
    FOREIGN_TABLE = 'f'
    PARTITIONED_TABLE = 'p'  # PostgresSQL 10+


pg_class = table(
    'pg_class',
    column('oid'),
    column('relname'),
    column('relacl'),
    column('relnamespace'),
    column('relkind'),
    column('relowner'),
)

pg_namespace = table(
    'pg_namespace',
    column('oid'),
    column('nspname'),
    column('nspowner'),
    column('nspacl'),
)

pg_roles = table(
    'pg_roles',
    column('oid'),
    column('rolname'),
)

pg_proc = table(
    'pg_proc',
    column('oid'),
    column('proname'),
    column('proargtypes'),
    column('pronamespace'),
    column('proacl'),
    column('proowner'),
)

pg_type = table(
    'pg_type',
    column('oid'),
    column('typname'),
    column('typnamespace'),
    column('typowner'),
    column('typacl'),
)

pg_language = table(
    'pg_language',
    column('oid'),
    column('lanname'),
    column('lanowner'),
    column('lanacl'),
)

pg_database = table(
    'pg_database',
    column('oid'),
    column('datname'),
    column('datdba'),
    column('datacl'),
)

pg_tablespace = table(
    'pg_tablespace',
    column('oid'),
    column('spcname'),
    column('spcowner'),
    column('spcacl'),
)

pg_attribute = table(
    'pg_attribute',
    column('attrelid'),
    column('attname'),
    column('attnum'),
    column('attisdropped'),
    column('attacl'),
)

_pg_class_stmt = (
    select(
        pg_class.c.oid,
        pg_namespace.c.nspname.label('schema'),
        pg_class.c.relname.label('name'),
        pg_roles.c.rolname.label('owner'),
        cast(pg_class.c.relacl, ARRAY(Text)).label('acl'),
    )
    .outerjoin(pg_namespace, pg_class.c.relnamespace == pg_namespace.c.oid)
    .outerjoin(pg_roles, pg_class.c.relowner == pg_roles.c.oid)
)

_pg_attribute_stmt = (
    select(
        pg_class.c.oid.label('table_oid'),
        pg_namespace.c.nspname.label('schema'),
        pg_class.c.relname.label('table'),
        pg_attribute.c.attname.label('column'),
        pg_roles.c.rolname.label('owner'),
        cast(pg_attribute.c.attacl, ARRAY(Text)).label('acl'),
    )
    .select_from(pg_attribute)
    .join(pg_class, pg_attribute.c.attrelid == pg_class.c.oid)
    .outerjoin(pg_namespace, pg_class.c.relnamespace == pg_namespace.c.oid)
    .outerjoin(pg_roles, pg_class.c.relowner == pg_roles.c.oid)
    .where(pg_attribute.c.attnum > 0)
    .where(~pg_attribute.c.attisdropped)
    .where(pg_class.c.relkind.in_([
        PgRelKind.TABLE.value,
        PgRelKind.VIEW.value,
        PgRelKind.MATERIALIZED_VIEW.value,
        PgRelKind.PARTITIONED_TABLE.value,
        PgRelKind.FOREIGN_TABLE.value,
    ]))
)

_upat = unnest(pg_proc.c.proargtypes).alias('upat')
_pg_proc_argtypes = (
    select(coalesce(array_agg(canonical_type(pg_type.c.typname)), cast([], ARRAY(Text))))
    .join(_upat, _upat.column == pg_type.c.oid)
    .scalar_subquery()
)

_pg_proc_stmt = (
    select(
        pg_proc.c.oid,
        pg_namespace.c.nspname.label('schema'),
        pg_proc.c.proname.label('name'),
        _pg_proc_argtypes.label('arg_types'),
        pg_roles.c.rolname.label('owner'),
        cast(pg_proc.c.proacl, ARRAY(Text)).label('acl'),
    )
    .outerjoin(pg_namespace, pg_proc.c.pronamespace == pg_namespace.c.oid)
    .outerjoin(pg_roles, pg_proc.c.proowner == pg_roles.c.oid)
)

_pg_lang_stmt = (
    select(
        pg_language.c.oid,
        pg_language.c.lanname.label('name'),
        pg_roles.c.rolname.label('owner'),
        cast(pg_language.c.lanacl, ARRAY(Text)).label('acl'),
    )
    .select_from(pg_language)
    .outerjoin(pg_roles, pg_language.c.lanowner == pg_roles.c.oid)
)

_pg_schema_stmt = (
    select(
        pg_namespace.c.oid,
        pg_namespace.c.nspname.label('name'),
        pg_roles.c.rolname.label('owner'),
        cast(pg_namespace.c.nspacl, ARRAY(Text)).label('acl'),
    )
    .select_from(pg_namespace)
    .outerjoin(pg_roles, pg_namespace.c.nspowner == pg_roles.c.oid)
)

_pg_db_stmt = (
    select(
        pg_database.c.oid,
        pg_database.c.datname.label('name'),
        pg_roles.c.rolname.label('owner'),
        cast(pg_database.c.datacl, ARRAY(Text)).label('acl'),
    )
    .select_from(pg_database)
    .outerjoin(pg_roles, pg_database.c.datdba == pg_roles.c.oid)
)

_pg_tablespace_stmt = (
    select(
        pg_tablespace.c.oid,
        pg_tablespace.c.spcname.label('name'),
        pg_roles.c.rolname.label('owner'),
        cast(pg_tablespace.c.spcacl, ARRAY(Text)).label('acl'),
    )
    .select_from(pg_tablespace)
    .outerjoin(pg_roles, pg_tablespace.c.spcowner == pg_roles.c.oid)
)

_pg_type_stmt = (
    select(
        pg_type.c.oid,
        pg_namespace.c.nspname.label('schema'),
        pg_type.c.typname.label('name'),
        pg_roles.c.rolname.label('owner'),
        cast(pg_type.c.typacl, ARRAY(Text)).label('acl'),
    )
    .select_from(pg_type)
    .outerjoin(pg_namespace, pg_type.c.typnamespace == pg_namespace.c.oid)
    .outerjoin(pg_roles, pg_type.c.typowner == pg_roles.c.oid)
)


def _filter_pg_class_stmt(stmt, schema=None, rel_name=None):
    if schema is not None:
        stmt = stmt.where(pg_namespace.c.nspname == schema)

    if rel_name is not None:
        if schema is None:
            # "pg_table_is_visible can also be used with views, materialized
            # views, indexes, sequences and foreign tables"
            stmt = stmt.where(pg_table_is_visible(pg_class.c.oid))

        stmt = stmt.where(pg_class.c.relname == rel_name)

    return stmt


def _filter_pg_proc_stmt(schema=None, function_name=None, arg_types=None):
    stmt = _pg_proc_stmt

    if (function_name is None) != (arg_types is None):
        raise TypeError(
            'function_name and arg_types must both be specified')  # pragma: no cover

    if schema is not None:
        stmt = stmt.where(pg_namespace.c.nspname == schema)

    if function_name is not None:
        arg_types = list(arg_types)

        if arg_types:
            arg_types_sub = (
                select(array_agg(canonical_type(column('typs'))))
                .select_from(func.unnest(arg_types).alias('typs'))
                .scalar_subquery()
            )
        else:
            arg_types_sub = cast([], ARRAY(Text))

        if schema is None:
            stmt = stmt.where(pg_function_is_visible(pg_proc.c.oid))
        stmt = stmt.where(pg_proc.c.proname == function_name)
        stmt = stmt.where(_pg_proc_argtypes == arg_types_sub)

    return stmt


def _filter_pg_type_stmt(schema=None, type_name=None):
    stmt = _pg_type_stmt

    if schema is not None:
        stmt = stmt.where(pg_namespace.c.nspname == schema)

    if type_name is not None:
        if schema is None:
            stmt = stmt.where(pg_type_is_visible(pg_type.c.oid))
        stmt = stmt.where(pg_type.c.typname == type_name)

    return stmt


def _table_stmt(schema=None, table_name=None):
    stmt = _filter_pg_class_stmt(
        _pg_class_stmt, schema=schema, rel_name=table_name)
    return stmt.where(pg_class.c.relkind.in_([
        PgRelKind.TABLE.value,
        PgRelKind.VIEW.value,
        PgRelKind.MATERIALIZED_VIEW.value,
        PgRelKind.PARTITIONED_TABLE.value,
        PgRelKind.FOREIGN_TABLE.value,
    ]))


def _sequence_stmt(schema=None, sequence_name=None):
    stmt = _filter_pg_class_stmt(
        _pg_class_stmt, schema=schema, rel_name=sequence_name)
    return stmt.where(pg_class.c.relkind == PgRelKind.SEQUENCE.value)


def get_all_table_acls(conn, schema=None):
    """Get privileges for all tables, views, materialized views, and foreign
    tables.

    Specify `schema` to limit the results to that schema.

    Returns:
        List of :class:`~.types.SchemaRelationInfo` objects.
    """
    stmt = _table_stmt(schema=schema)
    return [SchemaRelationInfo(**row) for row in conn.execute(stmt).mappings()]


def get_table_acl(conn, name, schema=None):
    """Get privileges for the table, view, materialized view, or foreign table
    specified by `name`.

    If `schema` is not given, the table or view must be visible in the search
    path.

    Returns:
         :class:`~.types.SchemaRelationInfo`
    """
    stmt = _table_stmt(schema=schema, table_name=name)
    row = conn.execute(stmt).mappings().first()
    if row is None:
        raise NoSuchObjectError(name)
    return SchemaRelationInfo(**row)


def get_all_column_acls(conn, schema=None):
    """Get privileges for all table, view, materialized view, and foreign
    table columns.

    Specify `schema` to limit the results to that schema.

    Returns:
        List of :class:`~.types.ColumnInfo` objects.
    """
    stmt = _filter_pg_class_stmt(_pg_attribute_stmt, schema=schema)
    return [ColumnInfo(**row) for row in conn.execute(stmt).mappings()]


def get_column_acls(conn, table_name, schema=None):
    """Get column privileges for the table, view, materialized view, or foreign
    table specified by `name`.

    If `schema` is not given, the table or view must be visible in the search
    path.

    Returns:
         List of :class:`~.types.ColumnInfo` objects.
    """
    stmt = _filter_pg_class_stmt(
        _pg_attribute_stmt, schema=schema, rel_name=table_name)
    rows = conn.execute(stmt).mappings().all()
    if not rows:
        raise NoSuchObjectError(table_name)
    return [ColumnInfo(**row) for row in rows]


def get_all_sequence_acls(conn, schema=None):
    """Unless `schema` is given, returns all sequences from all schemas.

    Returns:
        List of :class:`~.types.SchemaRelationInfo` objects.
    """
    stmt = _sequence_stmt(schema=schema)
    return [SchemaRelationInfo(**row) for row in conn.execute(stmt).mappings()]


def get_sequence_acl(conn, sequence, schema=None):
    """If `schema` is not given, the sequence must be visible in the search
    path.

    Returns:
         :class:`~.types.SchemaRelationInfo`
    """
    stmt = _sequence_stmt(schema=schema, sequence_name=sequence)
    row = conn.execute(stmt).mappings().first()
    if row is None:
        raise NoSuchObjectError(sequence)
    return SchemaRelationInfo(**row)


def _make_canonical_type_function(conn):
    """Create function which canonicalizes a type name, returning the input
    when casting to REGTYPE fails.

    E.g. casting the 'any' type fails. Normal examples include 'int4' -> 'integer'
    """
    # pg_temp is per-connection
    stmt = text("""
        CREATE OR REPLACE FUNCTION pg_temp.pg_grant_canonical_type(typname text)
        RETURNS text AS $$
        BEGIN
          BEGIN
            typname := typname::regtype::text;
          EXCEPTION WHEN syntax_error THEN
          END;
          RETURN typname;
        END;
        $$
        LANGUAGE plpgsql
        STABLE
        RETURNS NULL ON NULL INPUT;
    """)
    conn.execute(stmt)


def get_all_function_acls(conn, schema=None):
    """Unless `schema` is given, returns all functions from all schemas.

    Returns:
        List of :class:`~.types.FunctionInfo` objects.
    """
    _make_canonical_type_function(conn)
    stmt = _filter_pg_proc_stmt(schema=schema)
    return [FunctionInfo(**row) for row in conn.execute(stmt).mappings()]


def get_function_acl(conn, function_name, arg_types: Sequence[str], schema=None):
    """If `schema` is not given, the function must be visible in the search path.

    Returns:
         :class:`~.types.FunctionInfo`
    """
    # We could ask the user to register an event on their connection pool which
    # creates this function on checkout, but that isn't a nice API for the
    # common case.
    _make_canonical_type_function(conn)

    if (function_name is None) != (arg_types is None):
        raise TypeError('function_name and arg_types must both be specified')

    if not isinstance(arg_types, Sequence) or isinstance(arg_types, str):
        raise TypeError("arg_types should be a sequence of strings, e.g. ['text']")

    stmt = _filter_pg_proc_stmt(schema, function_name, arg_types)
    row = conn.execute(stmt).mappings().first()
    if row is None:
        raise NoSuchObjectError(function_name)
    return FunctionInfo(**row)


def get_all_language_acls(conn):
    """
    Returns:
        List of :class:`~.types.RelationInfo` objects.
    """
    return [RelationInfo(**row) for row in conn.execute(_pg_lang_stmt).mappings()]


def get_language_acl(conn, language):
    """
    Returns:
         :class:`~.types.RelationInfo`
    """
    stmt = _pg_lang_stmt.where(pg_language.c.lanname == language)
    row = conn.execute(stmt).mappings().first()
    if row is None:
        raise NoSuchObjectError(language)
    return RelationInfo(**row)


def get_all_schema_acls(conn):
    """
    Returns:
        List of :class:`~.types.RelationInfo` objects.
    """
    return [RelationInfo(**row) for row in conn.execute(_pg_schema_stmt).mappings()]


def get_schema_acl(conn, schema):
    """
    Returns:
         :class:`~.types.RelationInfo`
    """
    stmt = _pg_schema_stmt.where(pg_namespace.c.nspname == schema)
    row = conn.execute(stmt).mappings().first()
    if row is None:
        raise NoSuchObjectError(schema)
    return RelationInfo(**row)


def get_all_database_acls(conn):
    """
    Returns:
        List of :class:`~.types.RelationInfo` objects.
    """
    return [RelationInfo(**row) for row in conn.execute(_pg_db_stmt).mappings()]


def get_database_acl(conn, database):
    """
    Returns:
         :class:`~.types.RelationInfo`
    """
    stmt = _pg_db_stmt.where(pg_database.c.datname == database)
    row = conn.execute(stmt).mappings().first()
    if row is None:
        raise NoSuchObjectError(database)
    return RelationInfo(**row)


def get_all_tablespace_acls(conn):
    """
    Returns:
        List of :class:`~.types.RelationInfo` objects.
    """
    return [RelationInfo(**row) for row in conn.execute(_pg_tablespace_stmt).mappings()]


def get_tablespace_acl(conn, tablespace):
    """
    Returns:
         :class:`~.types.RelationInfo`
    """
    stmt = _pg_tablespace_stmt.where(pg_tablespace.c.spcname == tablespace)
    row = conn.execute(stmt).mappings().first()
    if row is None:
        raise NoSuchObjectError(tablespace)
    return RelationInfo(**row)


def get_all_type_acls(conn, schema=None):
    """Unless `schema` is given, returns all types from all schemas.

    Returns:
        List of :class:`~.types.SchemaRelationInfo` objects.
    """
    stmt = _filter_pg_type_stmt(schema=schema)
    return [SchemaRelationInfo(**row) for row in conn.execute(stmt).mappings()]


def get_type_acl(conn, type_name, schema=None):
    """If `schema` is not given, the type must be visible in the search path.

    Returns:
         :class:`~.types.SchemaRelationInfo`
    """
    stmt = _filter_pg_type_stmt(schema=schema, type_name=type_name)
    row = conn.execute(stmt).mappings().first()
    if row is None:
        raise NoSuchObjectError(type_name)
    return SchemaRelationInfo(**row)



import pytest

from pg_grant import NoSuchObjectError
from pg_grant.query import get_all_table_acls, get_table_acl


expected_acls = {
    'public': {
        # table1 has default privileges, so None is returned.
        'table1': None,
        # alice is owner, bob was granted all
        'table2': {'alice=arwdDxt/alice', 'bob=ar*wdDxt/alice'},
        # view1 has default privileges, so None is returned.
        'view1': None,
        # alice is owner, bob was granted INSERT
        'view2': {'alice=arwdDxt/alice', 'bob=a/alice'},
        # mview1 has default privileges, so None is returned.
        'mview1': None,
    },
}


def as_set(v):
    if v is not None:
        return set(v)


@pytest.mark.parametrize('name, acls', expected_acls['public'].items())
def test_get_table_acl_visible(connection, name, acls):
    """Find visible (i.e. in search path) tables matching ``name``."""
    table = get_table_acl(connection, name)
    assert as_set(table.acl) == acls


@pytest.mark.parametrize('schema, name, acls', [
    (schema, name, acl)
    for schema, d in expected_acls.items()
    for name, acl in d.items()
])
def test_get_table_acl_schema(connection, schema, name, acls):
    """Find tables from ``schema`` matching ``name``."""
    table = get_table_acl(connection, name, schema)
    assert as_set(table.acl) == acls


def test_get_all_table_acls(connection):
    """Get all sequences in all schemas."""
    tables = get_all_table_acls(connection)
    schemas = {x.schema for x in tables}
    assert schemas == {'public', 'information_schema', 'pg_catalog'}

    tested = 0

    for table in tables:
        if table.schema not in expected_acls:
            continue

        if table.name not in expected_acls[table.schema]:
            continue

        assert as_set(table.acl) == expected_acls[table.schema][table.name]
        tested += 1

    assert tested == sum(len(v) for v in expected_acls.values())


def test_no_such_object(connection):
    with pytest.raises(NoSuchObjectError):
        get_table_acl(connection, 'table3')



import pytest

from pg_grant import NoSuchObjectError
from pg_grant.query import get_all_sequence_acls, get_sequence_acl


expected_acls = {
    'public': {
        # seq1 has default privileges, so None is returned.
        'seq1': None,
        # alice is owner, bob was granted all
        'seq2': ['alice=rwU/alice', 'bob=rwU/alice'],
    },
    'schema1': {
        'seq3': None,
    }
}


@pytest.mark.parametrize('name, acls', expected_acls['public'].items())
def test_get_sequence_acl_visible(connection, name, acls):
    """Find visible (i.e. in search path) sequences matching ``name``."""
    sequence = get_sequence_acl(connection, name)
    assert sequence.acl == acls


@pytest.mark.parametrize('schema, name, acls', [
    (schema, seq, acl)
    for schema, d in expected_acls.items()
    for seq, acl in d.items()
])
def test_get_sequence_acl_schema(connection, schema, name, acls):
    """Find sequences  from ``schema`` matching ``name``."""
    sequence = get_sequence_acl(connection, name, schema)
    assert sequence.acl == acls


def test_get_all_sequence_acls(connection):
    """Get all sequences in all schemas."""
    sequences = get_all_sequence_acls(connection)
    assert {x.schema for x in sequences} == {'public', 'schema1'}

    tested = 0

    for sequence in sequences:
        if sequence.schema not in expected_acls:
            continue

        if sequence.name not in expected_acls[sequence.schema]:
            continue

        assert sequence.acl == expected_acls[sequence.schema][sequence.name]
        tested += 1

    assert tested == sum(len(v) for v in expected_acls.values())


def test_no_such_object(connection):
    with pytest.raises(NoSuchObjectError):
        get_sequence_acl(connection, 'seq3')



import pytest

from pg_grant import NoSuchObjectError
from pg_grant.query import get_all_column_acls, get_column_acls


expected_acls = {
    'public': {
        'table1': {
            'id': None,
        },
        'table2': {
            'id': None,
            'user': {'charlie=r/alice'},
        },
        'view1': {
            'id': None,
        },
        'view2': {
            'id': {'charlie=arwx/alice'},
            'user': None,
        },
        'mview1': {
            'id': None,
        },
    },
}


def as_set(v):
    if v is not None:
        return set(v)


@pytest.mark.parametrize('table_name, columns', expected_acls['public'].items())
def test_get_column_acls_visible(connection, table_name, columns):
    """Find columns for visible (i.e. in search path) tables matching
    ``table_name``.
    """
    column_acls = get_column_acls(connection, table_name)
    column_acls = {r.column: r.acl for r in column_acls}
    assert column_acls.keys() == columns.keys()

    for col, acl in columns.items():
        assert as_set(column_acls[col]) == acl


@pytest.mark.parametrize('schema, table_name, columns', [
    (schema, table_name, columns)
    for schema, d in expected_acls.items()
    for table_name, columns in d.items()
])
def test_get_column_acls_schema(connection, schema, table_name, columns):
    """Find columns for tables from ``schema`` matching ``name``."""
    column_acls = get_column_acls(connection, table_name, schema)
    column_acls = {r.column: r.acl for r in column_acls}
    assert column_acls.keys() == columns.keys()

    for col, acl in columns.items():
        assert as_set(column_acls[col]) == acl


def test_get_all_column_acls(connection):
    """Get all sequences in all schemas."""
    column_acls = get_all_column_acls(connection)
    schemas = {x.schema for x in column_acls}
    assert schemas == {'public', 'information_schema', 'pg_catalog'}

    tested = 0

    for row in column_acls:
        if row.schema not in expected_acls:
            continue

        if row.table not in expected_acls[row.schema]:
            continue

        assert as_set(row.acl) == expected_acls[row.schema][row.table][row.column]
        tested += 1

    assert tested == sum(len(w) for v in expected_acls.values() for w in v.values())


def test_no_such_object(connection):
    with pytest.raises(NoSuchObjectError):
        get_column_acls(connection, 'table3')



import pytest

from pg_grant import PgObjectType, Privileges, get_default_privileges


@pytest.mark.parametrize('type, public_priv', [
    (PgObjectType.TABLE, None),
    (PgObjectType.SEQUENCE, None),
    (PgObjectType.FUNCTION, ['EXECUTE']),
    (PgObjectType.LANGUAGE, ['USAGE']),
    (PgObjectType.SCHEMA, None),
    (PgObjectType.DATABASE, ['CONNECT', 'TEMPORARY']),
    (PgObjectType.TABLESPACE, None),
    (PgObjectType.TYPE, ['USAGE']),
    (PgObjectType.FOREIGN_DATA_WRAPPER, None),
    (PgObjectType.FOREIGN_SERVER, None),
    (PgObjectType.FOREIGN_TABLE, None),
    (PgObjectType.LARGE_OBJECT, None),
])
def test_default_privileges(type, public_priv):
    owner = 'alice'
    owner_priv = Privileges(grantee=owner, grantor=owner, privs=['ALL'])

    expected = [owner_priv]

    if public_priv:
        expected.append(Privileges(grantee='PUBLIC', grantor=owner, privs=public_priv))

    assert get_default_privileges(type, owner) == expected



import pytest

from pg_grant import NoSuchObjectError
from pg_grant.query import get_all_database_acls, get_database_acl


expected_acls = {
    # postgres is owner, public get TEMPORARY and CONNECT,
    # alice was granted CONNECT
    'db1': ['=Tc/postgres', 'postgres=CTc/postgres', 'alice=c*/postgres'],
    # postgres has default privileges, so None is returned.
    'postgres': None,
}


@pytest.mark.parametrize('name, acls', expected_acls.items())
def test_get_database_acl(connection, name, acls):
    """Find visible databases matching ``name``."""
    database = get_database_acl(connection, name)
    assert database.acl == acls


def test_get_all_database_acls(connection):
    """Get all databases in all databases."""
    databases = get_all_database_acls(connection)

    tested = 0

    for database in databases:
        if database.name not in expected_acls:
            continue

        assert database.acl == expected_acls[database.name]
        tested += 1

    assert tested == len(expected_acls)


def test_no_such_object(connection):
    with pytest.raises(NoSuchObjectError):
        get_database_acl(connection, 'db2')



import pytest

from pg_grant import PgObjectType, Privileges, parse_acl, parse_acl_item

parse_data = [
    ('alice=arwdDxt/alice', 'alice', 'alice',
     ['SELECT', 'UPDATE', 'INSERT', 'DELETE', 'TRUNCATE', 'REFERENCES',
      'TRIGGER'], []),
    ('alice=a*r*w*d*D*x*t*/alice', 'alice', 'alice', [],
     ['SELECT', 'UPDATE', 'INSERT', 'DELETE', 'TRUNCATE', 'REFERENCES',
      'TRIGGER']),
    ('bob=arwdDxt/alice', 'bob', 'alice',
     ['SELECT', 'UPDATE', 'INSERT', 'DELETE', 'TRUNCATE', 'REFERENCES',
      'TRIGGER'], []),
    ('=arwdDx*t/alice', 'PUBLIC', 'alice',
     ['SELECT', 'UPDATE', 'INSERT', 'DELETE', 'TRUNCATE', 'TRIGGER'],
     ['REFERENCES']),
    ('"odd=name"=a/alice', 'odd=name', 'alice', ['INSERT'], []),
    ('"esc""ape"=a/alice', 'esc"ape', 'alice', ['INSERT'], []),

    ('bob=a/alice', 'bob', 'alice', ['INSERT'], []),
    ('bob=r/alice', 'bob', 'alice', ['SELECT'], []),
    ('bob=w/alice', 'bob', 'alice', ['UPDATE'], []),
    ('bob=d/alice', 'bob', 'alice', ['DELETE'], []),
    ('bob=D/alice', 'bob', 'alice', ['TRUNCATE'], []),
    ('bob=x/alice', 'bob', 'alice', ['REFERENCES'], []),
    ('bob=t/alice', 'bob', 'alice', ['TRIGGER'], []),
    ('bob=X/alice', 'bob', 'alice', ['EXECUTE'], []),
    ('bob=U/alice', 'bob', 'alice', ['USAGE'], []),
    ('bob=C/alice', 'bob', 'alice', ['CREATE'], []),
    ('bob=c/alice', 'bob', 'alice', ['CONNECT'], []),
    ('bob=T/alice', 'bob', 'alice', ['TEMPORARY'], []),

    ('bob=a*/alice', 'bob', 'alice', [], ['INSERT']),
    ('bob=r*/alice', 'bob', 'alice', [], ['SELECT']),
    ('bob=w*/alice', 'bob', 'alice', [], ['UPDATE']),
    ('bob=d*/alice', 'bob', 'alice', [], ['DELETE']),
    ('bob=D*/alice', 'bob', 'alice', [], ['TRUNCATE']),
    ('bob=x*/alice', 'bob', 'alice', [], ['REFERENCES']),
    ('bob=t*/alice', 'bob', 'alice', [], ['TRIGGER']),
    ('bob=X*/alice', 'bob', 'alice', [], ['EXECUTE']),
    ('bob=U*/alice', 'bob', 'alice', [], ['USAGE']),
    ('bob=C*/alice', 'bob', 'alice', [], ['CREATE']),
    ('bob=c*/alice', 'bob', 'alice', [], ['CONNECT']),
    ('bob=T*/alice', 'bob', 'alice', [], ['TEMPORARY']),
]

parse_table_data = [
    ('alice=arwdDxt/alice', 'alice', 'alice', ['ALL'], []),
    ('alice=a*r*w*d*D*x*t*/alice', 'alice', 'alice', [], ['ALL']),
    ('bob=arwdDxt/alice', 'bob', 'alice', ['ALL'], []),
    ('bob=arwdDx/alice', 'bob', 'alice',
     ['SELECT', 'UPDATE', 'INSERT', 'REFERENCES', 'DELETE', 'TRUNCATE'], []),

    ('bob=a/alice', 'bob', 'alice', ['INSERT'], []),
    ('bob=r/alice', 'bob', 'alice', ['SELECT'], []),
    ('bob=w/alice', 'bob', 'alice', ['UPDATE'], []),
    ('bob=d/alice', 'bob', 'alice', ['DELETE'], []),
    ('bob=D/alice', 'bob', 'alice', ['TRUNCATE'], []),
    ('bob=x/alice', 'bob', 'alice', ['REFERENCES'], []),
    ('bob=t/alice', 'bob', 'alice', ['TRIGGER'], []),
    ('bob=X/alice', 'bob', 'alice', [], []),
    ('bob=U/alice', 'bob', 'alice', [], []),
    ('bob=C/alice', 'bob', 'alice', [], []),
    ('bob=c/alice', 'bob', 'alice', [], []),
    ('bob=T/alice', 'bob', 'alice', [], []),

    ('bob=a*/alice', 'bob', 'alice', [], ['INSERT']),
    ('bob=r*/alice', 'bob', 'alice', [], ['SELECT']),
    ('bob=w*/alice', 'bob', 'alice', [], ['UPDATE']),
    ('bob=d*/alice', 'bob', 'alice', [], ['DELETE']),
    ('bob=D*/alice', 'bob', 'alice', [], ['TRUNCATE']),
    ('bob=x*/alice', 'bob', 'alice', [], ['REFERENCES']),
    ('bob=t*/alice', 'bob', 'alice', [], ['TRIGGER']),
    ('bob=X*/alice', 'bob', 'alice', [], []),
    ('bob=U*/alice', 'bob', 'alice', [], []),
    ('bob=C*/alice', 'bob', 'alice', [], []),
    ('bob=c*/alice', 'bob', 'alice', [], []),
    ('bob=T*/alice', 'bob', 'alice', [], []),
]

parse_sequence_data = [
    ('alice=rwU/alice', 'alice', 'alice', ['ALL'], []),
    ('alice=r*w*U*/alice', 'alice', 'alice', [], ['ALL']),
    ('bob=rwU/alice', 'bob', 'alice', ['ALL'], []),
    ('bob=rw/alice', 'bob', 'alice', ['SELECT', 'UPDATE'], []),

    ('bob=a/alice', 'bob', 'alice', [], []),
    ('bob=r/alice', 'bob', 'alice', ['SELECT'], []),
    ('bob=w/alice', 'bob', 'alice', ['UPDATE'], []),
    ('bob=d/alice', 'bob', 'alice', [], []),
    ('bob=D/alice', 'bob', 'alice', [], []),
    ('bob=x/alice', 'bob', 'alice', [], []),
    ('bob=t/alice', 'bob', 'alice', [], []),
    ('bob=X/alice', 'bob', 'alice', [], []),
    ('bob=U/alice', 'bob', 'alice', ['USAGE'], []),
    ('bob=C/alice', 'bob', 'alice', [], []),
    ('bob=c/alice', 'bob', 'alice', [], []),
    ('bob=T/alice', 'bob', 'alice', [], []),

    ('bob=a*/alice', 'bob', 'alice', [], []),
    ('bob=r*/alice', 'bob', 'alice', [], ['SELECT']),
    ('bob=w*/alice', 'bob', 'alice', [], ['UPDATE']),
    ('bob=d*/alice', 'bob', 'alice', [], []),
    ('bob=D*/alice', 'bob', 'alice', [], []),
    ('bob=x*/alice', 'bob', 'alice', [], []),
    ('bob=t*/alice', 'bob', 'alice', [], []),
    ('bob=X*/alice', 'bob', 'alice', [], []),
    ('bob=U*/alice', 'bob', 'alice', [], ['USAGE']),
    ('bob=C*/alice', 'bob', 'alice', [], []),
    ('bob=c*/alice', 'bob', 'alice', [], []),
    ('bob=T*/alice', 'bob', 'alice', [], []),
]

parse_function_data = [
    ('alice=X/alice', 'alice', 'alice', ['ALL'], []),
    ('alice=X*/alice', 'alice', 'alice', [], ['ALL']),
    ('bob=X/alice', 'bob', 'alice', ['ALL'], []),

    ('bob=a/alice', 'bob', 'alice', [], []),
    ('bob=r/alice', 'bob', 'alice', [], []),
    ('bob=w/alice', 'bob', 'alice', [], []),
    ('bob=d/alice', 'bob', 'alice', [], []),
    ('bob=D/alice', 'bob', 'alice', [], []),
    ('bob=x/alice', 'bob', 'alice', [], []),
    ('bob=t/alice', 'bob', 'alice', [], []),
    ('bob=X/alice', 'bob', 'alice', ['ALL'], []),
    ('bob=U/alice', 'bob', 'alice', [], []),
    ('bob=C/alice', 'bob', 'alice', [], []),
    ('bob=c/alice', 'bob', 'alice', [], []),
    ('bob=T/alice', 'bob', 'alice', [], []),

    ('bob=a*/alice', 'bob', 'alice', [], []),
    ('bob=r*/alice', 'bob', 'alice', [], []),
    ('bob=w*/alice', 'bob', 'alice', [], []),
    ('bob=d*/alice', 'bob', 'alice', [], []),
    ('bob=D*/alice', 'bob', 'alice', [], []),
    ('bob=x*/alice', 'bob', 'alice', [], []),
    ('bob=t*/alice', 'bob', 'alice', [], []),
    ('bob=X*/alice', 'bob', 'alice', [], ['ALL']),
    ('bob=U*/alice', 'bob', 'alice', [], []),
    ('bob=C*/alice', 'bob', 'alice', [], []),
    ('bob=c*/alice', 'bob', 'alice', [], []),
    ('bob=T*/alice', 'bob', 'alice', [], []),
]

parse_language_data = [
    ('alice=U/alice', 'alice', 'alice', ['ALL'], []),
    ('alice=U*/alice', 'alice', 'alice', [], ['ALL']),
    ('bob=U/alice', 'bob', 'alice', ['ALL'], []),

    ('bob=a/alice', 'bob', 'alice', [], []),
    ('bob=r/alice', 'bob', 'alice', [], []),
    ('bob=w/alice', 'bob', 'alice', [], []),
    ('bob=d/alice', 'bob', 'alice', [], []),
    ('bob=D/alice', 'bob', 'alice', [], []),
    ('bob=x/alice', 'bob', 'alice', [], []),
    ('bob=t/alice', 'bob', 'alice', [], []),
    ('bob=X/alice', 'bob', 'alice', [], []),
    ('bob=U/alice', 'bob', 'alice', ['ALL'], []),
    ('bob=C/alice', 'bob', 'alice', [], []),
    ('bob=c/alice', 'bob', 'alice', [], []),
    ('bob=T/alice', 'bob', 'alice', [], []),

    ('bob=a*/alice', 'bob', 'alice', [], []),
    ('bob=r*/alice', 'bob', 'alice', [], []),
    ('bob=w*/alice', 'bob', 'alice', [], []),
    ('bob=d*/alice', 'bob', 'alice', [], []),
    ('bob=D*/alice', 'bob', 'alice', [], []),
    ('bob=x*/alice', 'bob', 'alice', [], []),
    ('bob=t*/alice', 'bob', 'alice', [], []),
    ('bob=X*/alice', 'bob', 'alice', [], []),
    ('bob=U*/alice', 'bob', 'alice', [], ['ALL']),
    ('bob=C*/alice', 'bob', 'alice', [], []),
    ('bob=c*/alice', 'bob', 'alice', [], []),
    ('bob=T*/alice', 'bob', 'alice', [], []),
]

parse_schema_data = [
    ('alice=CU/alice', 'alice', 'alice', ['ALL'], []),
    ('alice=C*U*/alice', 'alice', 'alice', [], ['ALL']),
    ('bob=CU/alice', 'bob', 'alice', ['ALL'], []),
    ('bob=C/alice', 'bob', 'alice', ['CREATE'], []),

    ('bob=a/alice', 'bob', 'alice', [], []),
    ('bob=r/alice', 'bob', 'alice', [], []),
    ('bob=w/alice', 'bob', 'alice', [], []),
    ('bob=d/alice', 'bob', 'alice', [], []),
    ('bob=D/alice', 'bob', 'alice', [], []),
    ('bob=x/alice', 'bob', 'alice', [], []),
    ('bob=t/alice', 'bob', 'alice', [], []),
    ('bob=X/alice', 'bob', 'alice', [], []),
    ('bob=U/alice', 'bob', 'alice', ['USAGE'], []),
    ('bob=C/alice', 'bob', 'alice', ['CREATE'], []),
    ('bob=c/alice', 'bob', 'alice', [], []),
    ('bob=T/alice', 'bob', 'alice', [], []),

    ('bob=a*/alice', 'bob', 'alice', [], []),
    ('bob=r*/alice', 'bob', 'alice', [], []),
    ('bob=w*/alice', 'bob', 'alice', [], []),
    ('bob=d*/alice', 'bob', 'alice', [], []),
    ('bob=D*/alice', 'bob', 'alice', [], []),
    ('bob=x*/alice', 'bob', 'alice', [], []),
    ('bob=t*/alice', 'bob', 'alice', [], []),
    ('bob=X*/alice', 'bob', 'alice', [], []),
    ('bob=U*/alice', 'bob', 'alice', [], ['USAGE']),
    ('bob=C*/alice', 'bob', 'alice', [], ['CREATE']),
    ('bob=c*/alice', 'bob', 'alice', [], []),
    ('bob=T*/alice', 'bob', 'alice', [], []),
]

parse_database_data = [
    ('alice=CcT/alice', 'alice', 'alice', ['ALL'], []),
    ('alice=C*c*T*/alice', 'alice', 'alice', [], ['ALL']),
    ('bob=CcT/alice', 'bob', 'alice', ['ALL'], []),
    ('bob=Cc/alice', 'bob', 'alice', ['CREATE', 'CONNECT'], []),

    ('bob=a/alice', 'bob', 'alice', [], []),
    ('bob=r/alice', 'bob', 'alice', [], []),
    ('bob=w/alice', 'bob', 'alice', [], []),
    ('bob=d/alice', 'bob', 'alice', [], []),
    ('bob=D/alice', 'bob', 'alice', [], []),
    ('bob=x/alice', 'bob', 'alice', [], []),
    ('bob=t/alice', 'bob', 'alice', [], []),
    ('bob=X/alice', 'bob', 'alice', [], []),
    ('bob=U/alice', 'bob', 'alice', [], []),
    ('bob=C/alice', 'bob', 'alice', ['CREATE'], []),
    ('bob=c/alice', 'bob', 'alice', ['CONNECT'], []),
    ('bob=T/alice', 'bob', 'alice', ['TEMPORARY'], []),

    ('bob=a*/alice', 'bob', 'alice', [], []),
    ('bob=r*/alice', 'bob', 'alice', [], []),
    ('bob=w*/alice', 'bob', 'alice', [], []),
    ('bob=d*/alice', 'bob', 'alice', [], []),
    ('bob=D*/alice', 'bob', 'alice', [], []),
    ('bob=x*/alice', 'bob', 'alice', [], []),
    ('bob=t*/alice', 'bob', 'alice', [], []),
    ('bob=X*/alice', 'bob', 'alice', [], []),
    ('bob=U*/alice', 'bob', 'alice', [], []),
    ('bob=C*/alice', 'bob', 'alice', [], ['CREATE']),
    ('bob=c*/alice', 'bob', 'alice', [], ['CONNECT']),
    ('bob=T*/alice', 'bob', 'alice', [], ['TEMPORARY']),
]

parse_tablespace_data = [
    ('alice=C/alice', 'alice', 'alice', ['ALL'], []),
    ('alice=C*/alice', 'alice', 'alice', [], ['ALL']),
    ('bob=C/alice', 'bob', 'alice', ['ALL'], []),

    ('bob=a/alice', 'bob', 'alice', [], []),
    ('bob=r/alice', 'bob', 'alice', [], []),
    ('bob=w/alice', 'bob', 'alice', [], []),
    ('bob=d/alice', 'bob', 'alice', [], []),
    ('bob=D/alice', 'bob', 'alice', [], []),
    ('bob=x/alice', 'bob', 'alice', [], []),
    ('bob=t/alice', 'bob', 'alice', [], []),
    ('bob=X/alice', 'bob', 'alice', [], []),
    ('bob=U/alice', 'bob', 'alice', [], []),
    ('bob=C/alice', 'bob', 'alice', ['ALL'], []),
    ('bob=c/alice', 'bob', 'alice', [], []),
    ('bob=T/alice', 'bob', 'alice', [], []),

    ('bob=a*/alice', 'bob', 'alice', [], []),
    ('bob=r*/alice', 'bob', 'alice', [], []),
    ('bob=w*/alice', 'bob', 'alice', [], []),
    ('bob=d*/alice', 'bob', 'alice', [], []),
    ('bob=D*/alice', 'bob', 'alice', [], []),
    ('bob=x*/alice', 'bob', 'alice', [], []),
    ('bob=t*/alice', 'bob', 'alice', [], []),
    ('bob=X*/alice', 'bob', 'alice', [], []),
    ('bob=U*/alice', 'bob', 'alice', [], []),
    ('bob=C*/alice', 'bob', 'alice', [], ['ALL']),
    ('bob=c*/alice', 'bob', 'alice', [], []),
    ('bob=T*/alice', 'bob', 'alice', [], []),
]

parse_type_data = parse_language_data
parse_fdw_data = parse_language_data
parse_foreign_server_data = parse_language_data

parse_foreign_table_data = [
    ('alice=r/alice', 'alice', 'alice', ['ALL'], []),
    ('alice=r*/alice', 'alice', 'alice', [], ['ALL']),
    ('bob=r/alice', 'bob', 'alice', ['ALL'], []),

    ('bob=a/alice', 'bob', 'alice', [], []),
    ('bob=r/alice', 'bob', 'alice', ['ALL'], []),
    ('bob=w/alice', 'bob', 'alice', [], []),
    ('bob=d/alice', 'bob', 'alice', [], []),
    ('bob=D/alice', 'bob', 'alice', [], []),
    ('bob=x/alice', 'bob', 'alice', [], []),
    ('bob=t/alice', 'bob', 'alice', [], []),
    ('bob=X/alice', 'bob', 'alice', [], []),
    ('bob=U/alice', 'bob', 'alice', [], []),
    ('bob=C/alice', 'bob', 'alice', [], []),
    ('bob=c/alice', 'bob', 'alice', [], []),
    ('bob=T/alice', 'bob', 'alice', [], []),

    ('bob=a*/alice', 'bob', 'alice', [], []),
    ('bob=r*/alice', 'bob', 'alice', [], ['ALL']),
    ('bob=w*/alice', 'bob', 'alice', [], []),
    ('bob=d*/alice', 'bob', 'alice', [], []),
    ('bob=D*/alice', 'bob', 'alice', [], []),
    ('bob=x*/alice', 'bob', 'alice', [], []),
    ('bob=t*/alice', 'bob', 'alice', [], []),
    ('bob=X*/alice', 'bob', 'alice', [], []),
    ('bob=U*/alice', 'bob', 'alice', [], []),
    ('bob=C*/alice', 'bob', 'alice', [], []),
    ('bob=c*/alice', 'bob', 'alice', [], []),
    ('bob=T*/alice', 'bob', 'alice', [], []),
]

parse_lob_data = [
    ('alice=rw/alice', 'alice', 'alice', ['ALL'], []),
    ('alice=r*w*/alice', 'alice', 'alice', [], ['ALL']),
    ('bob=rw/alice', 'bob', 'alice', ['ALL'], []),
    ('bob=r/alice', 'bob', 'alice', ['SELECT'], []),

    ('bob=a/alice', 'bob', 'alice', [], []),
    ('bob=r/alice', 'bob', 'alice', ['SELECT'], []),
    ('bob=w/alice', 'bob', 'alice', ['UPDATE'], []),
    ('bob=d/alice', 'bob', 'alice', [], []),
    ('bob=D/alice', 'bob', 'alice', [], []),
    ('bob=x/alice', 'bob', 'alice', [], []),
    ('bob=t/alice', 'bob', 'alice', [], []),
    ('bob=X/alice', 'bob', 'alice', [], []),
    ('bob=U/alice', 'bob', 'alice', [], []),
    ('bob=C/alice', 'bob', 'alice', [], []),
    ('bob=c/alice', 'bob', 'alice', [], []),
    ('bob=T/alice', 'bob', 'alice', [], []),

    ('bob=a*/alice', 'bob', 'alice', [], []),
    ('bob=r*/alice', 'bob', 'alice', [], ['SELECT']),
    ('bob=w*/alice', 'bob', 'alice', [], ['UPDATE']),
    ('bob=d*/alice', 'bob', 'alice', [], []),
    ('bob=D*/alice', 'bob', 'alice', [], []),
    ('bob=x*/alice', 'bob', 'alice', [], []),
    ('bob=t*/alice', 'bob', 'alice', [], []),
    ('bob=X*/alice', 'bob', 'alice', [], []),
    ('bob=U*/alice', 'bob', 'alice', [], []),
    ('bob=C*/alice', 'bob', 'alice', [], []),
    ('bob=c*/alice', 'bob', 'alice', [], []),
    ('bob=T*/alice', 'bob', 'alice', [], []),
]

parse_col_data = [
    ('alice=rwax/alice', 'alice', 'alice', ['ALL (id)'], []),
    ('alice=r*w*a*x*/alice', 'alice', 'alice', [], ['ALL (id)']),
    ('bob=rwax/alice', 'bob', 'alice', ['ALL (id)'], []),
    ('bob=rwa/alice', 'bob', 'alice', ['SELECT (id)', 'UPDATE (id)', 'INSERT (id)'], []),

    ('bob=a/alice', 'bob', 'alice', ['INSERT (id)'], []),
    ('bob=r/alice', 'bob', 'alice', ['SELECT (id)'], []),
    ('bob=w/alice', 'bob', 'alice', ['UPDATE (id)'], []),
    ('bob=d/alice', 'bob', 'alice', [], []),
    ('bob=D/alice', 'bob', 'alice', [], []),
    ('bob=x/alice', 'bob', 'alice', ['REFERENCES (id)'], []),
    ('bob=t/alice', 'bob', 'alice', [], []),
    ('bob=X/alice', 'bob', 'alice', [], []),
    ('bob=U/alice', 'bob', 'alice', [], []),
    ('bob=C/alice', 'bob', 'alice', [], []),
    ('bob=c/alice', 'bob', 'alice', [], []),
    ('bob=T/alice', 'bob', 'alice', [], []),

    ('bob=a*/alice', 'bob', 'alice', [], ['INSERT (id)']),
    ('bob=r*/alice', 'bob', 'alice', [], ['SELECT (id)']),
    ('bob=w*/alice', 'bob', 'alice', [], ['UPDATE (id)']),
    ('bob=d*/alice', 'bob', 'alice', [], []),
    ('bob=D*/alice', 'bob', 'alice', [], []),
    ('bob=x*/alice', 'bob', 'alice', [], ['REFERENCES (id)']),
    ('bob=t*/alice', 'bob', 'alice', [], []),
    ('bob=X*/alice', 'bob', 'alice', [], []),
    ('bob=U*/alice', 'bob', 'alice', [], []),
    ('bob=C*/alice', 'bob', 'alice', [], []),
    ('bob=c*/alice', 'bob', 'alice', [], []),
    ('bob=T*/alice', 'bob', 'alice', [], []),
]


@pytest.mark.parametrize('acl, grantee, grantor, privs, privswgo', parse_data)
def test_parse(acl, grantee, grantor, privs, privswgo):
    assert parse_acl_item(acl) == Privileges(grantee, grantor, privs, privswgo)


def test_parse_acl():
    assert parse_acl(['ali=a/ali', 'bob=a/ali']) == [
        Privileges(grantee='ali', grantor='ali', privs=['INSERT'], privswgo=[]),
        Privileges(grantee='bob', grantor='ali', privs=['INSERT'], privswgo=[])
    ]


@pytest.mark.parametrize(
    'acl, grantee, grantor, privs, privswgo', parse_table_data)
def test_parse_table(acl, grantee, grantor, privs, privswgo):
    parsed = parse_acl_item(acl, PgObjectType.TABLE)
    assert parsed == Privileges(grantee, grantor, privs, privswgo)


@pytest.mark.parametrize(
    'acl, grantee, grantor, privs, privswgo', parse_sequence_data)
def test_parse_sequence(acl, grantee, grantor, privs, privswgo):
    parsed = parse_acl_item(acl, PgObjectType.SEQUENCE)
    assert parsed == Privileges(grantee, grantor, privs, privswgo)


@pytest.mark.parametrize(
    'acl, grantee, grantor, privs, privswgo', parse_function_data)
def test_parse_function(acl, grantee, grantor, privs, privswgo):
    parsed = parse_acl_item(acl, PgObjectType.FUNCTION)
    assert parsed == Privileges(grantee, grantor, privs, privswgo)


@pytest.mark.parametrize(
    'acl, grantee, grantor, privs, privswgo', parse_language_data)
def test_parse_language(acl, grantee, grantor, privs, privswgo):
    parsed = parse_acl_item(acl, PgObjectType.LANGUAGE)
    assert parsed == Privileges(grantee, grantor, privs, privswgo)


@pytest.mark.parametrize(
    'acl, grantee, grantor, privs, privswgo', parse_schema_data)
def test_parse_schema(acl, grantee, grantor, privs, privswgo):
    parsed = parse_acl_item(acl, PgObjectType.SCHEMA)
    assert parsed == Privileges(grantee, grantor, privs, privswgo)


@pytest.mark.parametrize(
    'acl, grantee, grantor, privs, privswgo', parse_database_data)
def test_parse_database(acl, grantee, grantor, privs, privswgo):
    parsed = parse_acl_item(acl, PgObjectType.DATABASE)
    assert parsed == Privileges(grantee, grantor, privs, privswgo)


@pytest.mark.parametrize(
    'acl, grantee, grantor, privs, privswgo', parse_tablespace_data)
def test_parse_tablespace(acl, grantee, grantor, privs, privswgo):
    parsed = parse_acl_item(acl, PgObjectType.TABLESPACE)
    assert parsed == Privileges(grantee, grantor, privs, privswgo)


@pytest.mark.parametrize(
    'acl, grantee, grantor, privs, privswgo', parse_type_data)
def test_parse_type(acl, grantee, grantor, privs, privswgo):
    parsed = parse_acl_item(acl, PgObjectType.TYPE)
    assert parsed == Privileges(grantee, grantor, privs, privswgo)


@pytest.mark.parametrize(
    'acl, grantee, grantor, privs, privswgo', parse_fdw_data)
def test_parse_fdw(acl, grantee, grantor, privs, privswgo):
    parsed = parse_acl_item(acl, PgObjectType.FOREIGN_DATA_WRAPPER)
    assert parsed == Privileges(grantee, grantor, privs, privswgo)


@pytest.mark.parametrize(
    'acl, grantee, grantor, privs, privswgo', parse_foreign_server_data)
def test_parse_foreign_server_data(acl, grantee, grantor, privs, privswgo):
    parsed = parse_acl_item(acl, PgObjectType.FOREIGN_SERVER)
    assert parsed == Privileges(grantee, grantor, privs, privswgo)


@pytest.mark.parametrize(
    'acl, grantee, grantor, privs, privswgo', parse_foreign_table_data)
def test_parse_foreign_table_data(acl, grantee, grantor, privs, privswgo):
    parsed = parse_acl_item(acl, PgObjectType.FOREIGN_TABLE)
    assert parsed == Privileges(grantee, grantor, privs, privswgo)


@pytest.mark.parametrize(
    'acl, grantee, grantor, privs, privswgo', parse_lob_data)
def test_parse_lob_data(acl, grantee, grantor, privs, privswgo):
    parsed = parse_acl_item(acl, PgObjectType.LARGE_OBJECT)
    assert parsed == Privileges(grantee, grantor, privs, privswgo)


@pytest.mark.parametrize(
    'acl, grantee, grantor, privs, privswgo', parse_col_data)
def test_parse_col_data(acl, grantee, grantor, privs, privswgo):
    parsed = parse_acl_item(acl, PgObjectType.TABLE, 'id')
    assert parsed == Privileges(grantee, grantor, privs, privswgo)


@pytest.mark.parametrize('acl', [
    '"bob=a/alice',
    'bo"b=a/alice',
    '"bob""=a/alice',
    '"=a/alice',
    'bob=a/"alice',
    'bob=a/al"ice',
    'bob=a/"alice""',
    'bob=a/"',
])
def test_parse_invalid_quote(acl):
    with pytest.raises(ValueError) as exc_info:
        parse_acl_item(acl)

    assert 'quote' in exc_info.value.args[0]


def test_parse_unknown_type():
    with pytest.raises(ValueError) as exc_info:
        # noinspection PyTypeChecker
        parse_acl_item('bob=a/alice', 'bad type')

    assert 'Unknown type' in exc_info.value.args[0]



from typing import List, Optional, Sequence

from .types import Privileges, PgObjectType


def _get_acl_username(acl):
    """Port of ``copyAclUserName`` from ``dumputils.c``"""
    i = 0
    output = ''

    while i < len(acl) and acl[i] != '=':
        # If user name isn't quoted, then just add it to the output buffer
        if acl[i] != '"':
            output += acl[i]
            i += 1
        else:
            # Otherwise, it's a quoted username
            i += 1

            if i == len(acl):
                raise ValueError('ACL syntax error: unterminated quote.')

            # Loop until we come across an unescaped quote
            while not (acl[i] == '"' and acl[i + 1:i + 2] != '"'):
                # Quoting convention is to escape " as "".
                if acl[i] == '"' and acl[i + 1:i + 2] == '"':
                    i += 1

                output += acl[i]
                i += 1

                if i == len(acl):
                    raise ValueError('ACL syntax error: unterminated quote.')

            i += 1

    return i, output


def get_default_privileges(type: PgObjectType, owner: str) -> List[Privileges]:
    """Return a list of :class:`~pg_grant.types.Privileges` objects matching the
    default privileges for that type.

    This can be called when the ACL item from PostgreSQL is NULL to determine
    the implicit access privileges.

    .. seealso:: https://www.postgresql.org/docs/10/static/sql-grant.html
    """

    # "the owner has all privileges by default"
    # "PostgreSQL treats the owner's privileges as having been granted by the
    # owner to themselves"
    priv_list = [Privileges(grantee=owner, grantor=owner, privs=['ALL'])]

    public_privs = None

    # "PostgreSQL grants default privileges on some types of objects to PUBLIC.
    # No privileges are granted to PUBLIC by default on tables, columns,
    # schemas or tablespaces. For other types, the default privileges granted
    # to PUBLIC are as follows: CONNECT and CREATE TEMP TABLE for databases;
    # EXECUTE privilege for functions; and USAGE privilege for languages."
    if type is PgObjectType.DATABASE:
        public_privs = ['CONNECT', 'TEMPORARY']
    elif type is PgObjectType.FUNCTION:
        public_privs = ['EXECUTE']
    elif type is PgObjectType.LANGUAGE:
        public_privs = ['USAGE']
    elif type is PgObjectType.TYPE:
        # Seems like there's a documentation bug, and that types get USAGE by
        # default.
        # https://stackoverflow.com/questions/46656644
        public_privs = ['USAGE']

    if public_privs:
        priv_list.append(
            Privileges(grantee='PUBLIC', grantor=owner, privs=public_privs))

    return priv_list


def parse_acl(
    acl: Sequence[str],
    type: Optional[PgObjectType] = None,
    subname: Optional[str] = None,
) -> List[Privileges]:
    """

    Parameters:
        acl: ACL, e.g. ``['alice=arwdDxt/alice', 'bob=arwdDxt/alice']``
        type: Optional. If passed, all privileges may be reduced to ``['ALL']``.
        subname: Optional, e.g. for column privileges.

    Returns:
        List of :class:`~.types.Privileges`.

    .. seealso::

        This is a simple wrapper; :func:`.parse_acl_item` is called for each
        item in `acl`.
    """
    return [parse_acl_item(i, type, subname) for i in acl]


def parse_acl_item(
    acl_item: str,
    type: Optional[PgObjectType] = None,
    subname: Optional[str] = None,
) -> Privileges:
    """Port of ``parseAclItem`` from `dumputils.c`_

    Parameters:
        acl_item: ACL item, e.g. ``'alice=arwdDxt/bob'``
        type: Optional. If passed, all privileges may be reduced to ``['ALL']``.
        subname: Optional, e.g. for column privileges. Must be output from
                 :func:`psycopg2.extensions.quote_ident` or similar.

    .. warning::

        If the ``privs`` or ``privswgo`` attributes of the returned object will
        be used to construct an SQL statement, `subname` **must be a valid
        identifier** (e.g. by calling :func:`psycopg2.extensions.quote_ident`)
        in order to prevent SQL injection attacks.

        :func:`~pg_grant.sql.grant` and :func:`~pg_grant.sql.revoke` are not
        vulnerable, because those functions quote the embedded identifier:

        .. code-block:: python

            >>> from pg_grant import PgObjectType, parse_acl_item
            >>> from pg_grant.sql import grant
            >>> privs = parse_acl_item('alice=r/bob', subname='user')
            >>> privs.privs
            ['SELECT (user)']
            >>> str(grant(privs.privs, PgObjectType.TABLE, 'tbl1', 'alice'))
           'GRANT SELECT ("user") ON TABLE tbl1 TO alice'

        Note that ``"user"`` was quoted by :func:`~pg_grant.sql.grant`.

        In other cases, make sure to quote `subname`:

        .. code-block:: python

            >>> import psycopg2
            >>> from psycopg2.extensions import quote_ident
            >>> conn = psycopg2.connect(...)
            >>> parse_acl_item('alice=r/bob', subname=quote_ident('user', conn))
            >>> privs.privs
            ['SELECT ("user")']

    Returns:
        :class:`~.types.Privileges`
    """
    eq_pos, grantee = _get_acl_username(acl_item)
    assert acl_item[eq_pos] == '='

    if grantee == '':
        grantee = 'PUBLIC'

    slash_pos = acl_item.index('/', eq_pos)
    _, grantor = _get_acl_username(acl_item[slash_pos + 1:])

    privs = []
    privs_with_grant_option = []

    all_with_grant_option = all_without_grant_option = True

    priv = acl_item[eq_pos + 1:slash_pos]

    def convert_priv(code, keyword):
        nonlocal all_with_grant_option, all_without_grant_option

        pos = priv.find(code)
        if pos >= 0:
            if priv[pos + 1:pos + 2] == '*':
                s = keyword
                if subname is not None:
                    s += ' ({})'.format(subname)

                privs_with_grant_option.append(s)
                all_without_grant_option = False
            else:
                s = keyword
                if subname is not None:
                    s += ' ({})'.format(subname)

                privs.append(s)
                all_with_grant_option = False
        else:
            all_with_grant_option = all_without_grant_option = False

    if type is None:
        convert_priv('r', 'SELECT')
        convert_priv('w', 'UPDATE')
        convert_priv('a', 'INSERT')
        convert_priv('d', 'DELETE')
        convert_priv('D', 'TRUNCATE')
        convert_priv('x', 'REFERENCES')
        convert_priv('t', 'TRIGGER')
        convert_priv('X', 'EXECUTE')
        convert_priv('U', 'USAGE')
        convert_priv('C', 'CREATE')
        convert_priv('c', 'CONNECT')
        convert_priv('T', 'TEMPORARY')

        # Don't think anything can have all of them, but set all to False
        # since we don't know type.
        all_with_grant_option = all_without_grant_option = False
    elif type in {PgObjectType.TABLE, PgObjectType.SEQUENCE}:
        convert_priv('r', 'SELECT')
        convert_priv('w', 'UPDATE')

        if type is PgObjectType.SEQUENCE:
            convert_priv('U', 'USAGE')
        else:
            convert_priv('a', 'INSERT')
            convert_priv('x', 'REFERENCES')

            if subname is None:
                convert_priv('d', 'DELETE')
                convert_priv('t', 'TRIGGER')
                convert_priv('D', 'TRUNCATE')

    elif type is PgObjectType.FUNCTION:
        convert_priv('X', 'EXECUTE')
    elif type is PgObjectType.LANGUAGE:
        convert_priv('U', 'USAGE')
    elif type is PgObjectType.SCHEMA:
        convert_priv('C', 'CREATE')
        convert_priv('U', 'USAGE')
    elif type is PgObjectType.DATABASE:
        convert_priv('C', 'CREATE')
        convert_priv('c', 'CONNECT')
        convert_priv('T', 'TEMPORARY')
    elif type is PgObjectType.TABLESPACE:
        convert_priv('C', 'CREATE')
    elif type is PgObjectType.TYPE:
        convert_priv('U', 'USAGE')
    elif type is PgObjectType.FOREIGN_DATA_WRAPPER:
        convert_priv('U', 'USAGE')
    elif type is PgObjectType.FOREIGN_SERVER:
        convert_priv('U', 'USAGE')
    elif type is PgObjectType.FOREIGN_TABLE:
        convert_priv('r', 'SELECT')
    elif type is PgObjectType.LARGE_OBJECT:
        convert_priv('r', 'SELECT')
        convert_priv('w', 'UPDATE')
    else:
        raise ValueError('Unknown type: {}'.format(type))

    if all_with_grant_option:
        privs = []
        privs_with_grant_option = []
        s = 'ALL'

        if subname is not None:
            s += ' ({})'.format(subname)

        privs_with_grant_option.append(s)
    elif all_without_grant_option:
        privs = []
        privs_with_grant_option = []
        s = 'ALL'

        if subname is not None:
            s += ' ({})'.format(subname)

        privs.append(s)

    return Privileges(grantee, grantor, privs, privs_with_grant_option)



import re

from setuptools import setup, find_packages


INIT_FILE = 'pg_grant/__init__.py'
init_data = open(INIT_FILE).read()

metadata = dict(re.findall("__([a-z]+)__ = '([^']+)'", init_data))

VERSION = metadata['version']
LICENSE = metadata['license']
DESCRIPTION = metadata['description']
AUTHOR = metadata['author']
EMAIL = metadata['email']

requires = [
    'attrs',
]

extras_require = {
    ':python_version<"3.5"': [
        'typing',
    ],
    'sqlalchemy': [
        'sqlalchemy[postgresql]>=1.4',
    ],
    'test': [
        'plumbum',
        'pytest>=3.0',
    ],
    'docstest': [
        'doc8',
        'sphinx',
        'sphinx_rtd_theme',
    ],
    'pep8test': [
        'flake8',
        'pep8-naming',
    ],
}

setup(
    name='pg_grant',
    version=VERSION,
    description=DESCRIPTION,
    long_description=open('README.rst').read(),
    author=AUTHOR,
    author_email=EMAIL,
    url='https://github.com/RazerM/pg_grant',
    packages=find_packages(exclude=['tests']),
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    license=LICENSE,
    install_requires=requires,
    extras_require=extras_require,
    tests_require=extras_require['test'],
    python_requires='>=3.6',
)



import re

from sqlalchemy import inspect
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql.expression import ClauseElement, Executable

from pg_grant.types import PgObjectType

__all__ = (
    'grant',
    'revoke',
)

_re_valid_priv = re.compile(
    r'(SELECT|UPDATE|INSERT|DELETE|TRUNCATE|REFERENCES|TRIGGER|EXECUTE|USAGE'
    r'|CREATE|CONNECT|TEMPORARY|ALL)(?:\s+\((.*)\))?')


def _as_table(element):
    """Allow a Table or ORM model to be used as a table name."""
    insp = inspect(element, raiseerr=False)

    try:
        return insp.selectable
    except AttributeError:
        raise ValueError('Expected table element.')


class _GrantRevoke(Executable, ClauseElement):
    valid_privileges = {
        'SELECT',
        'UPDATE',
        'INSERT',
        'DELETE',
        'TRUNCATE',
        'REFERENCES',
        'TRIGGER',
        'EXECUTE',
        'USAGE',
        'CREATE',
        'CONNECT',
        'TEMPORARY',
        'ALL',
    }
    keyword = None

    def __init__(self, privileges, type: PgObjectType, target, grantee,
                 grant_option=False, schema=None, arg_types=None,
                 quote_subname=True):

        if privileges == 'ALL':
            privileges = ['ALL']

        self.privileges = privileges
        self.priv_type = type
        self.target = target
        self.grantee = grantee
        self.grant_option = grant_option
        self.schema = schema
        self.arg_types = arg_types
        self.quote_subname = quote_subname


class _Grant(_GrantRevoke):
    inherit_cache = False

    keyword = 'GRANT'


class _Revoke(_GrantRevoke):
    inherit_cache = False

    keyword = 'REVOKE'


@compiles(_GrantRevoke)
def _pg_grant(element, compiler, **kw):
    target = element.target
    schema = element.schema
    arg_types = element.arg_types
    priv_type = element.priv_type

    preparer = compiler.preparer

    privs = []

    for priv in element.privileges:
        match = _re_valid_priv.match(priv)
        if match is None:
            raise ValueError('Privilege not valid: {}'.format(priv))

        subname = match.group(2)

        if subname is not None:
            if element.quote_subname:
                subname = preparer.quote(subname)

            privs.append('{} ({})'.format(match.group(1), subname))
        else:
            privs.append(match.group(1))

    priv = ', '.join(privs)

    str_target = None

    if isinstance(target, str):
        if schema is not None:
            str_target = preparer.quote_schema(schema) + '.' + preparer.quote(target)
        else:
            str_target = preparer.quote(target)
    else:
        if schema is not None:
            raise ValueError('schema argument not supported unless target is a string.')

    if priv_type is not PgObjectType.FUNCTION and arg_types is not None:
        raise ValueError('arg_types argument not supported unless type is FUNCTION.')

    if priv_type is PgObjectType.TABLE:
        if str_target is not None:
            target = str_target
        else:
            target = compiler.process(_as_table(target), ashint=True)
    elif priv_type is PgObjectType.SEQUENCE:
        if str_target is not None:
            target = str_target
        else:
            target = preparer.format_sequence(target)
    elif priv_type is PgObjectType.TYPE:
        if str_target is None:
            target = compiler.process(target)
        else:
            target = str_target
    elif priv_type is PgObjectType.FUNCTION:
        if str_target is None:
            target = compiler.process(target)
        else:
            if arg_types is None:
                raise ValueError('Must use an empty sequence if function has '
                                 'no arguments, not None.')

            str_arg_types = ', '.join([preparer.quote(t) for t in arg_types])
            target = '{}({})'.format(str_target, str_arg_types)
    elif isinstance(priv_type, PgObjectType):
        if str_target is None:
            target = compiler.process(target)
        else:
            target = str_target
    else:
        raise ValueError('Unknown type: {}'.format(priv_type))

    is_grant = element.keyword == 'GRANT'

    grantee = element.grantee

    if grantee.upper() != 'PUBLIC':
        grantee = compiler.preparer.quote(element.grantee)

    return '{}{} {} ON {} {} {} {}{}'.format(
        element.keyword,
        ' GRANT OPTION FOR' if element.grant_option and not is_grant else '',
        priv,
        priv_type.value,
        target,
        'TO' if is_grant else 'FROM',
        grantee,
        ' WITH GRANT OPTION' if element.grant_option and is_grant else '',
    )


def grant(privileges, type: PgObjectType, target, grantee, grant_option=False,
          schema=None, arg_types=None, quote_subname=True):
    """GRANT statement that may be executed by SQLAlchemy.

    Parameters:
        privileges: List of privileges (or ``'ALL'``).
        type: PostgreSQL object type.
        target: Object name, or appropriate SQLAlchemy object (e.g.
                :class:`~sqlalchemy.schema.Table` or a declarative class).
        grantee: Role to receive privileges.
        grant_option: Whether the recipient may in turn grant these privileges
                      to others.
        schema: Optional schema, if `target` is a string.
        arg_types: Sequence of argument types for granting privileges on
                   functions. E.g. ``('int4', 'int4')`` or ``()``.
        quote_subname: Quote subname identifier in privileges, e.g.
                       ``'SELECT (user)'`` -> ``'SELECT ("user")``. This should
                       only be ``False`` if the subname is already a valid
                       identifier.

    .. seealso:: https://www.postgresql.org/docs/current/static/sql-grant.html
    """
    return _Grant(
        privileges, type, target, grantee, grant_option=grant_option,
        schema=schema, arg_types=arg_types, quote_subname=quote_subname)


def revoke(privileges, type: PgObjectType, target, grantee, grant_option=False,
           schema=None, arg_types=None, quote_subname=True):
    """REVOKE statement that may be executed by SQLAlchemy.

    Parameters:
        privileges: List of privileges (or ``'ALL'``).
        type: PostgreSQL object type.
        target: Object name, or appropriate SQLAlchemy object (e.g.
                :class:`~sqlalchemy.schema.Table` or a declarative class).
        grantee: Role to lose privileges.
        grant_option: Whether to revoke the grant option for these privileges.
        schema: Optional schema, if `target` is a string.
        arg_types: Sequence of argument types for revoking privileges on
                   functions. E.g. ``('int4', 'int4')`` or ``()``.
        quote_subname: Quote subname identifier in privileges, e.g.
                       ``'SELECT (user)'`` -> ``'SELECT ("user")``. This should
                       only be ``False`` if the subname is already a valid
                       identifier.

    .. warning:: When ``grant_option=True``, only the grant option is revoked,
                 not the privilege(s).

    .. seealso:: https://www.postgresql.org/docs/current/static/sql-revoke.html

    """
    return _Revoke(
        privileges, type, target, grantee, grant_option=grant_option,
        schema=schema, arg_types=arg_types, quote_subname=quote_subname)



import pytest

from pg_grant import NoSuchObjectError
from pg_grant.query import get_all_tablespace_acls, get_tablespace_acl


expected_acls = {
    # pg_default has default privileges, so None is returned.
    'pg_default': None,
    # postgres is owner, alice was granted CONNECT
    'pg_global': ['postgres=C/postgres', 'alice=C/postgres'],
}


@pytest.mark.parametrize('name, acls', expected_acls.items())
def test_get_tablespace_acl(connection, name, acls):
    """Find visible tablespaces matching ``name``."""
    tablespace = get_tablespace_acl(connection, name)
    assert tablespace.acl == acls


def test_get_all_tablespace_acls(connection):
    """Get all tablespaces in all tablespaces."""
    tablespaces = get_all_tablespace_acls(connection)

    tested = 0

    for tablespace in tablespaces:
        if tablespace.name not in expected_acls:
            continue

        assert tablespace.acl == expected_acls[tablespace.name]
        tested += 1

    assert tested == len(expected_acls)


def test_no_such_object(connection):
    with pytest.raises(NoSuchObjectError):
        get_tablespace_acl(connection, 'eggs')



from contextlib import suppress
from unittest.mock import Mock

import pytest
from sqlalchemy import func

from pg_grant import NoSuchObjectError
from pg_grant.query import (
    _make_canonical_type_function, get_all_function_acls, get_function_acl)

expected_acls = {
    'public': {
        # fun1 has default privileges, so None is returned.
        ('fun1', ('integer',)): None,
        # alice is owner, execute was revoked from public
        ('fun1', ('text',)): ['alice=X/alice'],
        ('fun2', ()): None,
    },
}

canonical_type = func.pg_temp.pg_grant_canonical_type


@pytest.mark.parametrize('signature, acls', expected_acls['public'].items())
def test_get_function_acl_visible(connection, signature, acls):
    """Find visible (i.e. in search path) functions matching ``name``."""
    name, arg_types = signature
    function = get_function_acl(connection, name, arg_types)
    assert function.acl == acls


def test_get_function_acl_canonical_type(connection):
    f1 = get_function_acl(connection, 'fun1', ('int4',), 'public')
    f2 = get_function_acl(connection, 'fun1', ('integer',), 'public')
    assert f1 == f2
    assert f1.arg_types == ['integer']


@pytest.mark.parametrize('type_name, canonical_name', [
    ('any', 'any'),
    ('anyarray', 'anyarray'),
    ('int4', 'integer'),
    ('float', 'double precision'),
    ('timestamptz', 'timestamp with time zone'),
])
def test_canonical_type_function(connection, type_name, canonical_name):
    _make_canonical_type_function(connection)
    assert connection.scalar(canonical_type(type_name)) == canonical_name


@pytest.mark.parametrize('schema, name, arg_types, acls', [
    (schema, name, arg_types, acl)
    for schema, d in expected_acls.items()
    for (name, arg_types), acl in d.items()
])
def test_get_function_acl_schema(connection, schema, name, arg_types, acls):
    """Find functions  from ``schema`` matching ``name``."""
    function = get_function_acl(connection, name, arg_types, schema)
    assert function.acl == acls


def test_get_all_function_acls(connection):
    """Get all sequences in all schemas."""
    functions = get_all_function_acls(connection)
    schemas = {x.schema for x in functions}
    assert schemas >= {'public', 'information_schema', 'pg_catalog'}

    tested = 0

    for function in functions:
        if function.schema not in expected_acls:
            continue

        key = (function.name, tuple(function.arg_types))
        print(key)

        if key not in expected_acls[function.schema]:
            continue

        assert function.acl == expected_acls[function.schema][key]
        tested += 1

    assert tested == sum(len(v) for v in expected_acls.values())


@pytest.mark.parametrize('function_name, arg_types', [
    (None, ('int4',)),
    ('fun1', None),
])
def test_missing_args(function_name, arg_types):
    with pytest.raises(TypeError) as exc_info:
        get_function_acl(Mock(), function_name, arg_types)

    msg = 'function_name and arg_types must both be specified'
    assert exc_info.value.args[0] == msg


@pytest.mark.parametrize('arg_types', [
    'a string',
    dict(),
    1
])
def test_invalid_arg_types_parameter(arg_types):
    with pytest.raises(TypeError) as exc_info:
        get_function_acl(Mock(), 'fun1', arg_types)

    msg = 'arg_types should be a sequence of strings'
    assert msg in exc_info.value.args[0]


@pytest.mark.parametrize('arg_types', [
    (),
    [],
    ['int4'],
    ('int4',),
])
def test_valid_arg_types_parameter(connection, arg_types):
    with suppress(NoSuchObjectError):
        get_function_acl(connection, 'fun1', arg_types)


def test_no_such_object(connection):
    with pytest.raises(NoSuchObjectError):
        get_function_acl(connection, 'fun3', [])
